# -*- coding: utf-8 -*-
"""
Helper functions that convert strftime formats into more readable
representations.
"""
import datetime
import re

from aiorest_ws.conf.global_settings import ISO_8601
from aiorest_ws.utils.date.formatters import iso8601_repr

__all__ = (
    'STRFDATETIME', 'STRFDATETIME_REPL',
    'date_formats', 'datetime_formats', 'time_formats',
    'humanize_strptime', 'humanize_timedelta',
)

STRFDATETIME = re.compile('([dgGhHis])')


def STRFDATETIME_REPL(match):
    return '%%(%s)s' % match.group()


def datetime_formats(formats):
    format = ', '.join(formats).replace(
        ISO_8601, 'YYYY-MM-DDThh:mm[:ss[.uuuuuu]][+HH:MM|-HH:MM|Z]'
    )
    return humanize_strptime(format)


def date_formats(formats):
    format = ', '.join(formats).replace(ISO_8601, 'YYYY[-MM[-DD]]')
    return humanize_strptime(format)


def time_formats(formats):
    format = ', '.join(formats).replace(ISO_8601, 'hh:mm[:ss[.uuuuuu]]')
    return humanize_strptime(format)


def humanize_strptime(format_string):
    # Note that we're missing some of the locale specific mappings that
    # don't really make sense
    mapping = {
        "%Y": "YYYY",
        "%y": "YY",
        "%m": "MM",
        "%b": "[Jan-Dec]",
        "%B": "[January-December]",
        "%d": "DD",
        "%H": "hh",
        "%I": "hh",  # Requires '%p' to differentiate from '%H'
        "%M": "mm",
        "%S": "ss",
        "%f": "uuuuuu",
        "%a": "[Mon-Sun]",
        "%A": "[Monday-Sunday]",
        "%p": "[AM|PM]",
        "%z": "[+HHMM|-HHMM]"
    }
    for key, val in mapping.items():
        format_string = format_string.replace(key, val)
    return format_string


def humanize_timedelta(timedelta, display="long", sep=", "):
    """
    Turns a datetime.timedelta object into a nice string representation.
    """
    assert isinstance(timedelta, datetime.timedelta), "First argument must be a timedelta."  # NOQA

    result = []

    weeks = int(timedelta.days / 7)
    days = timedelta.days % 7
    hours = int(timedelta.seconds / 3600)
    minutes = int((timedelta.seconds % 3600) / 60)
    seconds = timedelta.seconds % 60

    if display == "sql":
        days += weeks * 7
        return "%i %02i:%02i:%02i" % (days, hours, minutes, seconds)
    elif display == "iso8601":
        return iso8601_repr(timedelta)
    elif display == 'minimal':
        words = ["w", "d", "h", "m", "s"]
    elif display == 'short':
        words = [" wks", " days", " hrs", " min", " sec"]
    elif display == 'long':
        words = [" weeks", " days", " hours", " minutes", " seconds"]
    else:
        # Use django template-style formatting.
        # Valid values are: d,g,G,h,H,i,s
        return STRFDATETIME.sub(STRFDATETIME_REPL, display) % {
            'd': days,
            'g': hours,
            'G': hours if hours > 9 else '0%s' % hours,
            'h': hours,
            'H': hours if hours > 9 else '0%s' % hours,
            'i': minutes if minutes > 9 else '0%s' % minutes,
            's': seconds if seconds > 9 else '0%s' % seconds
        }

    values = [weeks, days, hours, minutes, seconds]

    for i in range(len(values)):
        if values[i]:
            if values[i] == 1 and len(words[i]) > 1:
                result.append("%i%s" % (values[i], words[i].rstrip('s')))
            else:
                result.append("%i%s" % (values[i], words[i]))

    # Values with less than one second, which are considered zeroes
    if len(result) == 0:
        # Display as 0 of the smallest unit
        result.append('0%s' % (words[-1]))

    return sep.join(result)



# -*- coding: utf-8 -*-
"""
XML classes and functions, used for serializing and de-serializing.
"""
from xml.sax.saxutils import XMLGenerator

__all__ = ('SimpleXMLGenerator', )


class SimpleXMLGenerator(XMLGenerator):
    """
    XML generator for input data.
    """
    root_name = 'root'
    item_tag_name = 'list-item'

    def __init__(self, stream, encoding='utf-8'):
        super(SimpleXMLGenerator, self).__init__(stream, encoding=encoding)

    def parse(self, data):
        """
        Convert data to XML.

        :param data: input data.
        """
        self.startDocument()
        self.startElement(self.root_name, {})
        self.to_xml(self, data)
        self.endElement(self.root_name)
        self.endDocument()

    def to_str(self, value):
        """
        Encode value for the string.

        :param value: value, which will have converted to the string.
        """
        return str(value).encode(self._encoding)

    def to_xml(self, xml, data):
        """
        Convert Python object to XML.

        :param xml: XML object.
        :param data: Python's object, which will have been converted.
        """
        if isinstance(data, (list, tuple)):
            for item in data:
                self.startElement(self.item_tag_name, {})
                self.to_xml(xml, item)
                self.endElement(self.item_tag_name)
        elif isinstance(data, dict):
            for key, value in data.items():
                xml.startElement(key, {})
                self.to_xml(xml, value)
                xml.endElement(key)
        elif data is None:
            pass
        else:
            xml.characters(self.to_str(data))



# -*- coding: utf-8 -*-
"""
This module contains classes and functions, used for configuration
issues with websockets.
"""
from autobahn.websocket.compress import PerMessageDeflateOffer, \
    PerMessageDeflateOfferAccept

__all__ = ('deflate_offer_accept', )


def deflate_offer_accept(offers):
    """
    Function to accept offers from the client.
    NOTE: For using this function you will need a "permessage-deflate"
    compression extension for WebSocket connections.

    :param offers: iterable object (list, tuple), where every object
                   is instance of PerMessageDeflateOffer.
    """
    for offer in offers:
        if isinstance(offer, PerMessageDeflateOffer):
            return PerMessageDeflateOfferAccept(offer)



# -*- coding: utf-8 -*-
__all__ = ('iso8601_repr', )


def iso8601_repr(timedelta, format=None):
    """
    Represent a timedelta as an ISO8601 duration.
    http://en.wikipedia.org/wiki/ISO_8601#Durations
    """
    years = int(timedelta.days / 365)
    weeks = int((timedelta.days % 365) / 7)
    days = timedelta.days % 7

    hours = int(timedelta.seconds / 3600)
    minutes = int((timedelta.seconds % 3600) / 60)
    seconds = timedelta.seconds % 60

    if format == 'alt':
        if years or weeks or days:
            raise ValueError('Does not support `alt` format for durations '
                             'more then 1 day')
        return 'PT{0:02d}:{1:02d}:{2:02d}'.format(hours, minutes, seconds)

    formatting = (
        ('P', (
            ('Y', years),
            ('W', weeks),
            ('D', days),
        )),
        ('T', (
            ('H', hours),
            ('M', minutes),
            ('S', seconds),
        )),
    )

    result = []
    for category, subcats in formatting:
        result += category
        for format, value in subcats:
            if value:
                result.append('%d%c' % (value, format))
    if result[-1] == 'T':
        result = result[:-1]

    return "".join(result)



# -*- coding: utf-8 -*-
"""
This module provide special classes and functions, which using for make
work with model fields more user-friendly.
"""
import collections
import inspect

__all__ = (
    'is_simple_callable', 'method_overridden', 'get_attribute', 'set_value',
    'to_choices_dict', 'flatten_choices_dict'
)


def is_simple_callable(obj):
    """
    True if the object is a callable that takes no arguments.
    """
    function = inspect.isfunction(obj)
    method = inspect.ismethod(obj)

    if not (function or method):
        return False

    args, _, _, defaults = inspect.getargspec(obj)
    len_args = len(args) if function else len(args) - 1
    len_defaults = len(defaults) if defaults else 0
    return len_args <= len_defaults


def method_overridden(method_name, cls, instance):
    """
    Determine if a method has been overridden.
    """
    method = getattr(cls, method_name)
    default_method = getattr(method, '__func__', method)
    return default_method is not getattr(instance, method_name).__func__


def get_attribute(instance, attrs):
    """
    Similar to Python's built in `getattr(instance, attr)`, but takes a list
    of nested attributes, instead of a single attribute. Also accepts either
    attribute lookup on objects or dictionary lookups.
    """
    for attr in attrs:
        if instance is None:
            # Break out early if we get `None` at any point in a nested lookup
            return None

        if isinstance(instance, collections.Mapping):
            instance = instance[attr]
        else:
            instance = getattr(instance, attr)

        if is_simple_callable(instance):
            try:
                instance = instance()
            except (AttributeError, KeyError) as exc:
                # If we raised an Attribute or KeyError here it'd get treated
                # as an omitted field in `Field.get_attribute()`. Instead we
                # raise a ValueError to ensure the exception is not masked
                raise ValueError(
                    'Exception raised in callable attribute "{0}"; '
                    'original exception was: {1}'.format(attr, exc)
                )

    return instance


def set_value(dictionary, keys, value):
    """
    Similar to Python's built in `dictionary[key] = value`,
    but takes a list of nested keys instead of a single key.
    set_value({'a': 1}, [], {'b': 2}) -> {'a': 1, 'b': 2}
    set_value({'a': 1}, ['x'], 2) -> {'a': 1, 'x': 2}
    set_value({'a': 1}, ['x', 'y'], 2) -> {'a': 1, 'x': {'y': 2}}
    """
    if not keys:
        dictionary.update(value)
        return

    for key in keys[:-1]:
        if key not in dictionary:
            dictionary[key] = {}
        dictionary = dictionary[key]

    dictionary[keys[-1]] = value


def to_choices_dict(choices):
    """
    Convert choices into key/value dicts.
    to_choices_dict([1]) -> {1: 1}
    to_choices_dict([(1, '1st'), (2, '2nd')]) -> {1: '1st', 2: '2nd'}
    to_choices_dict([('Group', ((1, '1'), 2))]) -> {'Group': {1: '1', 2: '2'}}
    """
    # Allow single, paired or grouped choices style:
    # choices = [1, 2, 3]
    # choices = [(1, 'First'), (2, 'Second'), (3, 'Third')]
    # choices = [('Category', ((1, 'First'), (2, 'Second'))), (3, 'Third')]
    ret = collections.OrderedDict()
    for choice in choices:
        if (not isinstance(choice, (list, tuple))):
            # Single choice
            ret[choice] = choice
        else:
            key, value = choice
            if isinstance(value, (list, tuple)):
                # Grouped choices (category, sub choices)
                ret[key] = to_choices_dict(value)
            else:
                # Paired choice (key, display value)
                ret[key] = value
    return ret


def flatten_choices_dict(choices):
    """
    Convert a group choices dict into a flat dict of choices.
    flatten_choices_dict({1: '1st', 2: '2nd'}) -> {1: '1st', 2: '2nd'}
    flatten_choices_dict({'Key': {1: '1st', 2: '2nd'}}) -> {1: '1st', 2: '2nd'}
    """
    ret = collections.OrderedDict()
    for key, value in choices.items():
        if isinstance(value, dict):
            # Grouped choices (category, sub choices)
            for sub_key, sub_value in value.items():
                ret[sub_key] = sub_value
        else:
            # Choice (key, display value)
            ret[key] = value
    return ret



# -*- coding: utf-8 -*-
"""
Classes and functions, which might be used for easier development purposes.
"""
__all__ = ('cached_property', )


class cached_property(object):
    """
    Decorator that transform a method with a single `self` argument into a
    property cached on the instance.
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls=None):
        result = instance.__dict__[self.func.__name__] = self.func(instance)
        return result



# -*- coding: utf-8 -*-
"""
Handled exceptions raised by aiorest-ws framework, which inspired under
Django REST framework.
"""
from aiorest_ws.status import WS_PROTOCOL_ERROR, WS_DATA_CANNOT_ACCEPT
from aiorest_ws.utils.encoding import force_text

__all__ = (
    'ImproperlyConfigured', 'BaseAPIException', 'EndpointValueError',
    'IncorrectArgument', 'IncorrectMethodNameType', 'InvalidHandler',
    'InvalidPathArgument', 'InvalidRenderer', 'NotImplementedMethod',
    'NotSpecifiedError', 'NotSpecifiedHandler', 'NotSpecifiedMethodName',
    'NotSpecifiedURL', 'NotSupportedArgumentType', 'SerializerError',
)


class ImproperlyConfigured(Exception):
    """
    Exception for the cases, when user has configured some class, setting or
    any other required parameter improperly.
    """
    pass


class BaseAPIException(Exception):
    """
    Base class for aiorest-ws framework exceptions.

    All subclasses should provide `.status_code` and `.default_detail`
    properties.
    """
    status_code = WS_PROTOCOL_ERROR
    default_detail = u"A server error occurred."

    def __init__(self, detail=None):
        """
        Create an instance of exception with users detail information if
        it is passed.

        :param detail: users detail information (string).
        """
        if detail is not None:
            self.detail = force_text(detail)
        else:
            self.detail = force_text(self.default_detail)

    def __str__(self):
        return self.detail


class EndpointValueError(BaseAPIException):
    default_detail = u"Incorrect endpoint. Check path to your API."


class IncorrectArgument(BaseAPIException):
    status_code = WS_DATA_CANNOT_ACCEPT
    default_detail = u"Check `args` in query on errors and try again."


class IncorrectMethodNameType(BaseAPIException):
    default_detail = u"Method name should be a string type."


class InvalidHandler(BaseAPIException):
    default_detail = u"Received handler isn't correct. It shall be function" \
                     u" or class, inherited from the MethodBasedView class."


class InvalidPathArgument(BaseAPIException):
    default_detail = u"Received path value not valid."


class InvalidRenderer(BaseAPIException):
    default_detail = "Attribute `renderers` should be defined as list or " \
                     "tuple of inherited from BaseSerializer classes."


class NotImplementedMethod(BaseAPIException):
    default_detail = u"Error occurred in not implemented method."


class NotSpecifiedError(BaseAPIException):
    default_detail = u"Not specified parameter."


class NotSpecifiedHandler(NotSpecifiedError):
    default_detail = u"For URL, typed in request, handler not specified."


class NotSpecifiedMethodName(NotSpecifiedError):
    default_detail = u"In query not specified `method` argument."


class NotSpecifiedURL(NotSpecifiedError):
    default_detail = u"In query not specified `url` argument."


class NotSupportedArgumentType(BaseAPIException):
    status_code = WS_DATA_CANNOT_ACCEPT
    default_detail = u"Check your arguments on supported types."


class SerializerError(BaseAPIException):
    default_detail = u"Error has occurred inside serializer class."



# -*- coding: utf-8 -*-
"""
We perform uniqueness checks explicitly on the serializer class, rather
the using Django's `.full_clean()`.

This gives us better separation of concerns, allows us to use single-step
object creation, and makes it possible to switch between using the implicit
`ModelSerializer` class and an equivalent explicit `Serializer` class.
"""
from django.db import DataError
from django.utils.translation import ugettext_lazy as _

from aiorest_ws.db.orm.exceptions import ValidationError
from aiorest_ws.utils.representation import smart_repr

__all__ = ('qs_exists', 'qs_filter', 'UniqueValidator', )


def qs_exists(queryset):
    try:
        return queryset.exists()
    except (TypeError, ValueError, DataError):
        return False


def qs_filter(queryset, **kwargs):
    try:
        return queryset.filter(**kwargs)
    except (TypeError, ValueError, DataError):
        return queryset.none()


class UniqueValidator(object):
    """
    Validator that corresponds to `unique=True` on a model field.
    Should be applied to an individual field on the serializer.
    """
    message = _('This field must be unique.')

    def __init__(self, queryset, message=None):
        self.queryset = queryset
        self.serializer_field = None
        self.message = message or self.message

    def set_context(self, serializer_field):
        """
        This hook is called by the serializer instance,
        prior to the validation call being made.
        """
        # Determine the underlying model field name. This may not be the
        # same as the serializer field name if `source=<>` is set
        self.field_name = serializer_field.source_attrs[-1]
        # Determine the existing instance, if this is an update operation
        self.instance = getattr(serializer_field.parent, 'instance', None)

    def filter_queryset(self, value, queryset):
        """
        Filter the queryset to all instances matching the given attribute.
        """
        filter_kwargs = {self.field_name: value}
        return qs_filter(queryset, **filter_kwargs)

    def exclude_current_instance(self, queryset):
        """
        If an instance is being updated, then do not include
        that instance itself as a uniqueness conflict.
        """
        if self.instance is not None:
            return queryset.exclude(pk=self.instance.pk)
        return queryset

    def __call__(self, value):
        queryset = self.queryset
        queryset = self.filter_queryset(value, queryset)
        queryset = self.exclude_current_instance(queryset)
        if qs_exists(queryset):
            raise ValidationError(self.message)

    def __repr__(self):
        return repr('<%s(queryset=%s)>' % (
            self.__class__.__name__,
            smart_repr(self.queryset)
        ))



# -*- coding: utf-8 -*-
"""
User abstractions for authentication.
"""
from aiorest_ws.auth.user.utils import generate_password_hash

__all__ = ('AbstractUser', 'User', )


class AbstractUser(object):
    """
    Abstract class, which describe generic user.
    """
    def __init__(self, *args, **kwargs):
        super(AbstractUser, self).__init__()
        self.first_name = kwargs.get('first_name', '')
        self.last_name = kwargs.get('last_name', '')
        self._change_user_credentials(*args, **kwargs)

    def _change_user_credentials(self, *args, **kwargs):
        """
        Change user credentials, based on the passed args and kwargs.
        """
        self._is_active = kwargs.get('is_active', True)
        self._is_superuser = kwargs.get('is_superuser', False)
        self._is_staff = kwargs.get('is_staff', False)
        self._is_user = kwargs.get('is_user', False)

        if any([self._is_superuser, self._is_staff, self._is_user]):
            self._is_anonymous = False
        else:
            self._is_anonymous = True

    def get_fullname(self):
        """
        Get fullname of user.
        """
        return u"{0} {1}".format(self.first_name, self.last_name).strip()

    @property
    def is_active(self):
        """
        Get is_active status of user.
        """
        return self._is_active

    @is_active.setter
    def is_active(self, is_active):
        """
        Set is_active status of user.

        :param is_active: boolean value of is_active.
        """
        self._is_active = is_active

    @property
    def is_superuser(self):
        """
        Get is_superuser status of user.
        """
        return self._is_superuser

    @property
    def is_staff(self):
        """
        Get is_staff status of user.
        """
        return self._is_staff

    @property
    def is_user(self):
        """
        Get is_user status of user.
        """
        return self._is_user

    @property
    def is_anonymous(self):
        """
        Get is_anonymous status of user.
        """
        return self._is_anonymous

    def is_authenticated(self):
        """
        Check that this user is authenticated.
        """
        if self.is_anonymous:
            return False
        else:
            return True


class User(AbstractUser):
    """
    Default class, which describe current user.
    """
    def __init__(self, *args, **kwargs):
        super(User, self).__init__(*args, **kwargs)
        self._id = kwargs.get('id', None)
        self._username = kwargs.get('username', '')
        self._password = kwargs.get('password', '')
        self._email = kwargs.get('email', '')
        self._permissions = kwargs.get('permissions', [])

    @property
    def id(self):
        """
        Get users ID.
        """
        return self._id

    @property
    def username(self):
        """
        Get username.
        """
        return self._username

    @username.setter
    def username(self, username):
        """
        Set username.

        :param username: new username as a string.
        """
        self._username = username

    @property
    def password(self):
        """
        Get password.
        """
        return self._password

    @password.setter
    def password(self, password):
        """
        Set new password for user.

        :param password: password as a string.
        """
        self._password = generate_password_hash(password)

    @property
    def email(self):
        """
        Get email.
        """
        return self._email

    @email.setter
    def email(self, email):
        """
        Set email for user.

        :param email: email as a string.
        """
        self._email = email

    @property
    def permissions(self):
        """
        Get list of permissions.
        """
        return self._permissions

    @permissions.setter
    def permissions(self, permissions):
        """
        Set permissions for user.

        :param permissions: list of permissions.
        """
        self._permissions = permissions

    def check_password(self, password):
        """
        Check for a valid password has taken.

        :param password: password as a string.
        """
        return self.password == generate_password_hash(password)

    def has_permission(self, obj=None):
        """
        Check that user have a some permission.

        :param obj: permissions object derived from the AbstractPermission.
        """
        if self.is_active:
            return obj in self.permissions
        else:
            return False



# -*- coding: utf-8 -*-
"""
Logging tool for aiorest-ws framework.
"""
import logging
import logging.config

from aiorest_ws.conf import settings

__all__ = ('logger', )

logging.config.dictConfig(settings.DEFAULT_LOGGING_SETTINGS)
logger = logging.getLogger('aiorest-ws')



# -*- coding: utf-8 -*-
"""
Decorators and wrappers, used for routing issues.
"""
from aiorest_ws.validators import MethodValidator
from aiorest_ws.views import MethodBasedView

__all__ = ('endpoint', )


def endpoint(path, methods, name=None, **attrs):
    """
    Decorator function, which turn handler into MethodBasedView class.

    :param path: URL, used for get access to APIs.
    :param methods: acceptable method name or list of methods.
    :param name: short name of endpoint.
    :param attrs: any other attributes, which must be initialized.
    """
    def endpoint_decorator(func):
        def wrapper():
            class FunctionView(MethodBasedView):
                def handler(self, request, *args, **kwargs):
                    return func(request, *args, **kwargs)

            view = FunctionView

            supported_methods = methods
            method_validator = MethodValidator()
            method_validator.validate(supported_methods)

            if type(supported_methods) is str:
                supported_methods = [supported_methods, ]

            for method in supported_methods:
                setattr(view, method.lower(), view.handler)

            for attr in attrs:
                setattr(view, str(attr).lower(), attrs[attr])

            return path, view, methods, name
        return wrapper
    return endpoint_decorator



# -*- coding: utf-8 -*-
"""
Functions and constants, which can be used for work with User models.
"""
from aiorest_ws.db.utils import convert_db_row_to_dict
from hashlib import sha256

SQL_CREATE_USER_TABLE = """
    CREATE TABLE IF NOT EXISTS aiorest_auth_user
    (id INTEGER PRIMARY KEY NOT NULL,
     username CHAR(255) NOT NULL,
     password CHAR(255) NOT NULL,
     last_name CHAR(255),
     first_name CHAR(255),
     is_active BOOL DEFAULT TRUE NOT NULL,
     is_superuser BOOL DEFAULT FALSE NOT NULL,
     is_staff BOOL DEFAULT FALSE NOT NULL,
     is_user BOOL DEFAULT TRUE NOT NULL
    );
"""
SQL_CREATE_TOKEN_FOREIGN_KEY = """
    ALTER TABLE aiorest_auth_token
    ADD COLUMN user_id INTEGER REFERENCES aiorest_auth_user(id);
"""
SQL_USER_ADD = """
    INSERT INTO aiorest_auth_user (`username`, `password`, `first_name`,
    `last_name`, `is_superuser`, `is_staff`, `is_user`, `is_active`)
    VALUES (?, ?, ?, ?, ?, ?, ?, ?);
"""
SQL_USER_GET = """
    SELECT `username`, `password`, `first_name`, `last_name`, `is_superuser`,
    `is_staff`, `is_user`, `is_active`
    FROM aiorest_auth_user
    WHERE id=?;
"""
SQL_USER_GET_WITH_ID = """
    SELECT `id`, `username`, `password`, `first_name`, `last_name`,
    `is_superuser`, `is_staff`, `is_user`, `is_active`
    FROM aiorest_auth_user
    WHERE username=?;
"""
SQL_USER_GET_BY_USERNAME = """
    SELECT `username`, `password`, `first_name`, `last_name`, `is_superuser`,
    `is_staff`, `is_user`, `is_active`
    FROM aiorest_auth_user
    WHERE username=?;
"""
SQL_USER_UPDATE = """
    UPDATE aiorest_auth_user
    SET {}
    WHERE username=?;
"""
USER_MODEL_FIELDS = (
    'id', 'username', 'password', 'first_name', 'last_name', 'is_superuser',
    'is_staff', 'is_user', 'is_active'
)
USER_MODEL_FIELDS_WITHOUT_PK = (
    'username', 'password', 'first_name', 'last_name', 'is_superuser',
    'is_staff', 'is_user', 'is_active'
)


def construct_update_sql(**parameters):
    """
    Create update SQL query for SQLite based on the data, provided by the user.

    :param parameters: dictionary, where key is updated field of user model.
    """
    query_args = []
    update_field_template = "{}=? "
    updated_fields = ''
    for index, (field, value) in enumerate(parameters.items()):
        updated_fields += update_field_template.format(field)

        if index < len(parameters) - 1:
            updated_fields += " ,"

        query_args.append(value)
    sql = SQL_USER_UPDATE.format(updated_fields)
    return sql, query_args


def convert_user_raw_data_to_dict(user_raw_data, with_id=False):
    """
    Convert database row to the dictionary object.

    :param user_raw_data: database row.
    :param with_id: boolean flag, which means necessity to append to the result
                    dictionary primary key of database row or not.
    """
    if with_id:
        fields_tuple = USER_MODEL_FIELDS
    else:
        fields_tuple = USER_MODEL_FIELDS_WITHOUT_PK

    bool_fields = ('is_superuser', 'is_staff', 'is_user', 'is_active')
    user_data = convert_db_row_to_dict(user_raw_data, fields_tuple)
    for field in bool_fields:
        user_data[field] = bool(user_data[field])
    return user_data


def generate_password_hash(password):
    """
    Generate SHA256 hash for password.

    :param password: password as a string.
    """
    return sha256(password.encode('utf-8')).hexdigest()



# -*- coding: utf-8 -*-
"""
Abstract classes for future implementation.
"""
from abc import ABCMeta, abstractmethod

__all__ = (
    'AbstractEndpoint', 'AbstractRouter', 'AbstractMiddleware',
    'AbstractPermission',
)


class AbstractEndpoint(metaclass=ABCMeta):
    """
    Base class for endpoints.
    """
    path = None      # URL, used for get access to API
    handler = None   # Class/function for processing request
    methods = []     # List of supported methods (GET, POST, etc.)
    name = None      # Short name for route
    _pattern = None  # Pattern, which using for checking path on compatible

    def __init__(self, path, handler, methods, name):
        self.path = path
        self.handler = handler
        if type(methods) is str:
            self.methods.append(methods)
        else:
            self.methods.extend(methods)
        self.name = name

    @abstractmethod
    def match(self, path):
        """
        Checking path on compatible.

        :param path: URL, which used for get access to API.
        """
        pass


class AbstractRouter(metaclass=ABCMeta):
    """
    Base class for routers.
    """
    _middlewares = []

    def __init__(self, *args, **kwargs):
        self._urls = []
        self._routes = {}

    @property
    def middlewares(self):
        """
        Get list of used middlewares.
        """
        return self._middlewares

    @abstractmethod
    def process_request(self, request):
        """
        Handling received request from user.

        :param request: request from user.
        """
        pass


class AbstractMiddleware(metaclass=ABCMeta):
    """
    Base class for middlewares.
    """
    @abstractmethod
    def process_request(self, request, handler):
        """
        Processing request before calling handler.

        :param request: instance of Request class.
        :param handler: view, invoked later for the request.
        """
        pass


class AbstractPermission(metaclass=ABCMeta):
    """
    Base class for permissions.
    """
    @staticmethod
    def check(request, handler):
        """
        Check permission method.

        :param request: instance of Request class.
        :param handler: view, invoked later for the request.
        """
        pass



# -*- coding: utf-8 -*-
"""
Module for initialize "global" settings and make it available from any part
of future application.

First of all, values will be read from user defined settings file, when path to
him is specified at AIOREST_SETTINGS_MODULE environment variable. Any other
settings, which not specified/defined by the user, will be taken from the
default aiorest_ws.conf.settings module.
"""
import importlib
import os

from aiorest_ws.conf import global_settings

__all__ = (
    'ENVIRONMENT_VARIABLE', 'Settings', 'UserSettingsHolder', 'settings',
)

ENVIRONMENT_VARIABLE = "AIORESTWS_SETTINGS_MODULE"


class Settings(object):
    """
    Settings class of application.
    """
    def __init__(self):
        super(Settings, self).__init__()

        # Parse default settings and append to object
        self._setup(global_settings)

        # After that take the user defined settings and make merge
        user_settings = os.environ.get(ENVIRONMENT_VARIABLE, None)
        if user_settings:
            user_settings = importlib.import_module(user_settings)
            self._setup(user_settings)

        self._create_database_managers()

    def _setup(self, settings):
        for setting in dir(settings):
            if setting.isupper():
                setattr(self, setting, getattr(settings, setting))

        self._wrapped = self

    def _create_database_managers(self):
        if not self.USE_ORM_ENGINE:

            for name, db_settings in self.DATABASES.items():
                backend = db_settings['backend']

                init_kwargs = ('name', 'user', 'password', 'host', 'port')
                kwargs = {
                    key: db_settings[key]
                    for key in db_settings if key in init_kwargs
                }
                managers_module = importlib.import_module(
                    backend + '.managers'
                )
                manager_cls = db_settings.get('manager') or 'SQLiteManager'
                # User defined manager_cls as string means that necessary to
                # extract class and create one instance of this
                if type(manager_cls) is str:
                    manager_instance = getattr(
                        managers_module, manager_cls
                    )(**kwargs)
                # In any other cases in can be already instantiated manager
                else:
                    manager_instance = manager_cls
                db_settings['manager'] = manager_instance


class UserSettingsHolder(Settings):
    """
    Holder for user configured settings.
    """
    # SETTINGS_MODULE doesn't make much sense in the manually configured
    # (standalone) case
    SETTINGS_MODULE = None

    def __init__(self, default_settings):
        """
        Requests for configuration variables not in this class are satisfied
        from the module specified in default_settings (if possible).
        """
        self.__dict__['_deleted'] = set()
        self.default_settings = default_settings

    def __getattr__(self, name):
        if name in self._deleted:
            raise AttributeError
        return getattr(self.default_settings, name)

    def __setattr__(self, name, value):
        self._deleted.discard(name)
        super(UserSettingsHolder, self).__setattr__(name, value)

    def __delattr__(self, name):
        self._deleted.add(name)
        if hasattr(self, name):
            super(UserSettingsHolder, self).__delattr__(name)

    def __dir__(self):
        return sorted(
            s for s in list(self.__dict__) + dir(self.default_settings)
            if s not in self._deleted
        )

    def is_overridden(self, setting):
        deleted = (setting in self._deleted)
        set_locally = (setting in self.__dict__)
        set_on_default = getattr(
            self.default_settings, 'is_overridden', lambda s: False
        )(setting)
        return (deleted or set_locally or set_on_default)

    def __repr__(self):
        return '<%(cls)s>' % {
            'cls': self.__class__.__name__,
        }


settings = Settings()



# -*- coding: utf-8 -*-
"""
Helper functions for mapping model fields to a dictionary of default
keyword arguments that should be used for their equivalent serializer fields.
"""
import inspect

from aiorest_ws.utils.text import capfirst

__all__ = ('ClassLookupDict', 'needs_label')


class ClassLookupDict(object):
    """
    Takes a dictionary with classes as keys.
    Lookups against this object will traverses the object's inheritance
    hierarchy in method resolution order, and returns the first matching value
    from the dictionary or raises a KeyError if nothing matches.
    """
    def __init__(self, mapping):
        self.mapping = mapping

    def __getitem__(self, key):
        if hasattr(key, '_proxy_class'):
            # Deal with proxy classes. Ie. BoundField behaves as if it
            # is a Field instance when using ClassLookupDict
            base_class = key._proxy_class
        else:
            base_class = key.__class__

        for cls in inspect.getmro(base_class):
            if cls in self.mapping:
                return self.mapping[cls]
        raise KeyError('Class %s not found in lookup.' % base_class.__name__)

    def __setitem__(self, key, value):
        self.mapping[key] = value


def needs_label(model_field_verbose_name, field_name):
    """
    Returns `True` if the label based on the model's verbose name is not equal
    to the default label it would have based on it's field name.
    """
    default_label = field_name.replace('_', ' ').capitalize()
    return capfirst(model_field_verbose_name) != default_label



# -*- coding: utf-8 -*-
"""
The `compat` module provides support for backwards compatibility with older
versions of Django, and compatibility wrappers around optional packages.
"""
import inspect

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.db import models


try:
    from django.contrib.postgres import fields as postgres_fields
except ImportError:
    postgres_fields = None


try:
    from django.contrib.postgres.fields import JSONField
except ImportError:
    JSONField = None


__all__ = [
    'postgres_fields', 'JSONField', '_resolve_model',
    'get_related_model', 'get_remote_field', 'value_from_object'
]


def _resolve_model(obj):
    """
    Resolve supplied `obj` to a Django model class.
    `obj` must be a Django model class itself, or a string
    representation of one.  Useful in situations like GH #1225 where
    Django may not have resolved a string-based reference to a model in
    another model's foreign key definition.
    String representations should have the format:
        'appname.ModelName'
    """
    if isinstance(obj, str) and len(obj.split('.')) == 2:
        app_name, model_name = obj.split('.')
        resolved_model = apps.get_model(app_name, model_name)
        if resolved_model is None:
            msg = "Django did not return a model for {0}.{1}"
            raise ImproperlyConfigured(msg.format(app_name, model_name))
        return resolved_model
    elif inspect.isclass(obj) and issubclass(obj, models.Model):
        return obj
    raise ValueError("{0} is not a Django model".format(obj))


def get_related_model(field):
    return field.remote_field.model


def get_remote_field(field, **kwargs):
    return field.remote_field


def value_from_object(field, obj):
    return field.value_from_object(obj)



# -*- coding: utf-8 -*-
"""
Classes and function for creating and processing requests from user.
"""
import asyncio
import json
from base64 import b64encode, b64decode

from autobahn.asyncio.websocket import WebSocketServerProtocol, \
    WebSocketServerFactory

from aiorest_ws.abstract import AbstractRouter
from aiorest_ws.routers import SimpleRouter
from aiorest_ws.validators import check_and_set_subclass
from aiorest_ws.wrappers import Request

__all__ = ('RequestHandlerProtocol', 'RequestHandlerFactory', )


class RequestHandlerProtocol(WebSocketServerProtocol):
    """
    REST WebSocket protocol instance, creating for every client connection.
    This protocol describe how to process network events (users requests to
    APIs) asynchronously.
    """
    def _decode_message(self, payload, isBinary=False):
        """
        Decoding input message to Request object.

        :param payload: input message.
        :param isBinary: boolean value, means that received data had a binary
                         format.
        """
        # Message was taken in base64
        if isBinary:
            payload = b64decode(payload)
        input_data = json.loads(payload.decode('utf-8'))
        return Request(**input_data)

    def _encode_message(self, response, isBinary=False):
        """
        Encoding output message.

        :param response: output message.
        :param isBinary: boolean value, means that received data had a binary
                         format.
        """
        # Encode additionally to base64 if necessary
        if isBinary:
            response = b64encode(response)
        return response

    @asyncio.coroutine
    def onMessage(self, payload, isBinary):
        """
        Handler, called for every message which was sent from the some user.

        :param payload: input message.
        :param isBinary: boolean value, means that received data had a binary
                         format.
        """
        request = self._decode_message(payload, isBinary)
        response = self.factory.router.process_request(request)
        out_payload = self._encode_message(response, isBinary)
        self.sendMessage(out_payload, isBinary=isBinary)


class RequestHandlerFactory(WebSocketServerFactory):
    """
    REST WebSocket server factory, which instantiates client connections.

    NOTE: Persistent configuration information is not saved in the instantiated
    protocol. For such cases kept data in a Factory classes, databases, etc.
    """
    def __init__(self, *args, **kwargs):
        super(RequestHandlerFactory, self).__init__(*args, **kwargs)
        self._router = kwargs.get('router', SimpleRouter(*args, **kwargs))

    @property
    def router(self):
        """
        Get router instance.
        """
        return self._router

    @router.setter
    def router(self, router):
        """
        Set router instance.

        :param router: instance of class, which derived from AbstractRouter.
        """
        if router:
            check_and_set_subclass(self, '_router', router, AbstractRouter)



# -*- coding: utf-8 -*-
"""
Functions and constants, which can be used for work with Token models.
"""

SQL_CREATE_TOKEN_TABLE = """
    CREATE TABLE IF NOT EXISTS aiorest_auth_token
    (id INTEGER PRIMARY KEY NOT NULL,
     name CHAR(64) NOT NULL, -- name of key (admin, api, etc.)
     token TEXT NOT NULL,
     created DATETIME DEFAULT CURRENT_TIMESTAMP NOT NULL,
     expired DATETIME DEFAULT NULL -- for some tokens it doesn't necessary
    );
"""
SQL_TOKEN_GET = """
    SELECT `id`, `name`, `token`, `created`, `expired`, `user_id`
    FROM aiorest_auth_token
    WHERE token=?;
"""
SQL_TOKEN_GET_BY_TOKEN_USERNAME = """
    SELECT aiorest_auth_token.id, `name`, `token`, `created`,
    `expired`, `user_id`
    FROM aiorest_auth_token
    INNER JOIN aiorest_auth_user
    ON aiorest_auth_token.user_id=aiorest_auth_user.id
    WHERE name=? AND username=?;
"""
SQL_TOKEN_ADD = """
    INSERT INTO aiorest_auth_token (`name`, `token`, `expired`, `user_id`)
    VALUES (?, ?, ?, ?);
"""
TOKEN_MODEL_FIELDS = ('id', 'name', 'token', 'created', 'expired', 'user_id')



# -*- coding: utf-8 -*-
"""
This module provides special classes (mixins) which used for getting data from
SQLAlchemy ORM when work with a model serializers or fields.
"""
from aiorest_ws.conf import settings
from aiorest_ws.db.orm.sqlalchemy.model_meta import model_pk


__all__ = ('ORMSessionMixin', 'SQLAlchemyMixin', )


class ORMSessionMixin(object):
    """
    Special wrapper around SQLALchemy querysets, when user specified them in
    for fields, serializers and etcetera.
    """
    def _get_session(self):
        return settings.SQLALCHEMY_SESSION()

    def get_queryset(self):
        """
        Return queryset, which will be executed during a session.
        Necessary to write queries for field in a models like:
            from sqlalchemy.orm.query import Query
            Query(User).filter(id==5).first()
        """
        # Avoid there "connection leaks", when developer described
        # queries like session.query(User).filter(id==5).first()
        if self.queryset.session:
            self.queryset.session.close()

        self.queryset.session = self._get_session()
        return self.queryset


class SQLAlchemyMixin(object):
    """
    Class which provide opportunity to get primary key from the passed object.
    """
    def _get_filter_args(self, query, data):
        mapper = query._bind_mapper()
        model = mapper.class_
        pk_fields = model_pk(model)
        return (getattr(model, field) == data[field] for field in pk_fields)

    def _get_object_pk(self, obj):
        mapper = obj.__mapper__
        model = mapper.class_
        pk_fields = model_pk(model)
        data = {str(field): getattr(obj, field) for field in pk_fields}
        return data if len(pk_fields) > 1 else data[pk_fields[0]]



# -*- coding: utf-8 -*-
"""
This module contains special classes, which using for making serializing
process more easier and better.

All this classes are based on serializers from Django REST framework. Original
code published under BSD license. For more details look into AUTHORS file.
"""
from collections import OrderedDict, MutableMapping

__all__ = (
    'ReturnDict', 'ReturnList', 'BoundField', 'NestedBoundField',
    'BindingDict',
)


class ReturnDict(OrderedDict):
    """
    Return object from `serialier.data` for the `Serializer` class.
    Includes a backlink to the serializer instance for renderers
    to use if they need richer field information.
    """

    def __init__(self, *args, **kwargs):
        self.serializer = kwargs.pop('serializer')
        super(ReturnDict, self).__init__(*args, **kwargs)

    def copy(self):
        return ReturnDict(self, serializer=self.serializer)

    def __repr__(self):
        return dict.__repr__(self)

    def __reduce__(self):
        # Pickling these objects will drop the .serializer backlink,
        # but preserve the raw data
        return dict, (dict(self),)


class ReturnList(list):
    """
    Return object from `serialier.data` for the `SerializerList` class.
    Includes a backlink to the serializer instance for renderers
    to use if they need richer field information.
    """

    def __init__(self, *args, **kwargs):
        self.serializer = kwargs.pop('serializer')
        super(ReturnList, self).__init__(*args, **kwargs)

    def __repr__(self):
        return list.__repr__(self)

    def __reduce__(self):
        # Pickling these objects will drop the .serializer backlink,
        # but preserve the raw data.
        return list, (list(self),)


class BoundField(object):
    """
    A field object that also includes `.value` and `.error` properties.
    Returned when iterating over a serializer instance,
    providing an API similar to Django forms and form fields.
    """

    def __init__(self, field, value, errors, prefix=''):
        self._field = field
        self._prefix = prefix
        self.value = value
        self.errors = errors
        self.name = prefix + self.field_name

    def __getattr__(self, attr_name):
        return getattr(self._field, attr_name)

    @property
    def _proxy_class(self):
        return self._field.__class__

    def __repr__(self):
        return u"<%s value=%s errors=%s>" % (
            self.__class__.__name__, self.value, self.errors
        )

    def as_form_field(self):
        value = '' if (self.value in (None, False)) else self.value
        return self.__class__(self._field, value, self.errors, self._prefix)


class NestedBoundField(BoundField):
    """
    This `BoundField` additionally implements __iter__ and __getitem__
    in order to support nested bound fields. This class is the type of
    `BoundField` that is used for serializer fields.
    """

    def __init__(self, field, value, errors, prefix=''):
        if value is None or value is '':
            value = {}
        super(NestedBoundField, self).__init__(field, value, errors, prefix)

    def __iter__(self):
        for field in self.fields.values():
            yield self[field.field_name]

    def __getitem__(self, key):
        field = self.fields[key]
        value = self.value.get(key) if self.value else None
        error = self.errors.get(key) if self.errors else None
        if hasattr(field, 'fields'):
            return NestedBoundField(
                field, value, error, prefix=self.name + '.'
            )
        return BoundField(field, value, error, prefix=self.name + '.')

    def as_form_field(self):
        values = {}
        for key, value in self.value.items():
            if isinstance(value, (list, dict)):
                values[key] = value
            else:
                values[key] = '' if (value in (None, False)) else value
        return self.__class__(self._field, values, self.errors, self._prefix)


class BindingDict(MutableMapping):
    """
    This dict-like object is used to store fields on a serializer.

    This ensures that whenever fields are added to the serializer we call
    `field.bind()` so that the `field_name` and `parent` attributes
    can be set correctly.
    """

    def __init__(self, serializer):
        self.serializer = serializer
        self.fields = OrderedDict()

    def __setitem__(self, key, field):
        self.fields[key] = field
        field.bind(field_name=key, parent=self.serializer)

    def __getitem__(self, key):
        return self.fields[key]

    def __delitem__(self, key):
        del self.fields[key]

    def __iter__(self):
        return iter(self.fields)

    def __len__(self):
        return len(self.fields)

    def __repr__(self):
        return dict.__repr__(self.fields)



# -*- coding: utf-8 -*-
"""
This module contains function, which using at validators.py for type
checking issues.
"""
from inspect import isclass

__all__ = ('to_str', 'get_object_type', )


def to_str(obj):
    """
    Custom convert object to string representation.
    """
    if isinstance(obj, (list, tuple)):
        string = "/".join([item.__name__ for item in obj])
    else:
        string = obj.__name__
    return string


def get_object_type(value):
    """
    Getting object type.
    """
    return type(value) if not isclass(value) else value



# -*- coding: utf-8 -*-
import os
import re
import ast
from setuptools import setup

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('aiorest_ws/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

requirements = ['autobahn==0.16.0', ]
test_requirements = requirements + [
    'pytest', 'pytest-asyncio',
    'pytest-cov', 'pytest-xdist'
]


def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, '__init__.py'))
    ]


args = dict(
    name='aiorest-ws',
    version=version,
    url='https://github.com/Relrin/aiorest-ws',
    license='BSD',
    author='Valeryi Savich',
    author_email='relrin78@gmail.com',
    description='REST framework with WebSockets support',
    packages=get_packages('aiorest_ws'),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=requirements,
    tests_require=test_requirements,
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP'
    ],
)


if __name__ == '__main__':
    setup(**args)



# -*- coding: utf-8 -*-
"""
Universal validators for ORM fields, which applied before commit into DB.

This gives us better separation of concerns, allows us to use single-step
object creation, and makes it possible to switch between using the implicit
`ModelSerializer` class and an equivalent explicit `Serializer` class.
"""
from aiorest_ws.db.orm.exceptions import ValidationError
from aiorest_ws.utils.representation import smart_repr

__all__ = (
    'BaseValidator', 'MaxValueValidator', 'MinValueValidator',
    'MaxLengthValidator', 'MinLengthValidator', 'EnumValidator',
    'BaseUniqueFieldValidator',
)


class BaseValidator(object):
    message = None
    code = None

    def __init__(self, *args, **kwargs):
        """
        Constructor for BaseValidator.

        :param message: string, which used as a message, when validation
        has fault.
        :param code: error code
        """
        self.message = kwargs.get('message', self.message)
        self.code = kwargs.get('code', self.code)

    def __call__(self, value):
        raise NotImplementedError('`__call__(self, value)` method must '
                                  'be implemented.')

    def __eq__(self, other):
        return (
            isinstance(other, BaseValidator) and
            (self.message == other.message) and
            (self.code == other.code)
        )


class MaxValueValidator(BaseValidator):
    """
    Validator for checking maximal value of input number.
    """
    message = u"Ensure this value is less than or equal to {max_value}."

    def __init__(self, max_value, *args, **kwargs):
        super(MaxValueValidator, self).__init__(*args, **kwargs)
        self.max_value = max_value
        if self.max_value is None:
            raise ValueError('Attribute `max_value` can not be NoneType.')

    def __call__(self, value):
        if value > self.max_value:
            message = self.message.format(max_value=self.max_value)
            raise ValidationError(message)


class MinValueValidator(BaseValidator):
    """
    Validator for checking minimal value of input number.
    """
    message = u"Ensure this value is greater than or equal to {min_value}."

    def __init__(self, min_value, *args, **kwargs):
        super(MinValueValidator, self).__init__(*args, **kwargs)
        self.min_value = min_value
        if self.min_value is None:
            raise ValueError('Attribute `min_value` can not be NoneType.')

    def __call__(self, value):
        if value < self.min_value:
            message = self.message.format(min_value=self.min_value)
            raise ValidationError(message)


class MaxLengthValidator(BaseValidator):
    """
    Validator for checking maximum length of input string.
    """
    message = u"Ensure that this value has no more {max_length} characters."

    def __init__(self, max_length, *args, **kwargs):
        super(MaxLengthValidator, self).__init__(*args, **kwargs)
        self.max_length = max_length
        if self.max_length and self.max_length < 0:
            raise ValueError('Attribute `max_length` can take only positive '
                             'numbers.')

    def __call__(self, value):
        if self.max_length and len(value) > self.max_length:
            message = self.message.format(max_length=self.max_length)
            raise ValidationError(message)


class MinLengthValidator(BaseValidator):
    """
    Validator for checking minimum length of input string.
    """
    message = u"Ensure that this value has minimum {min_length} characters."

    def __init__(self, min_length, *args, **kwargs):
        super(MinLengthValidator, self).__init__(*args, **kwargs)
        self.min_length = min_length
        if self.min_length and self.min_length < 0:
            raise ValueError('Attribute `min_length` can take only positive '
                             'numbers.')

    def __call__(self, value):
        if self.min_length and len(value) < self.min_length:
            message = self.message.format(min_length=self.min_length)
            raise ValidationError(message)


class EnumValidator(BaseValidator):
    """
    Validator for checking input keys of passed Enum.
    """
    message = u"Ensure that passed value is one of the allowable: {key_list}."

    def __init__(self, enum, *args, **kwargs):
        self.enum = enum
        super(EnumValidator, self).__init__(*args, **kwargs)

    @property
    def keys(self):
        return [member for member in self.enum.__members__]

    def is_enum_key(self, value):
        return value in self.keys

    def __call__(self, value):
        if not self.is_enum_key(value):
            message = self.message.format(key_list=self.keys)
            raise ValidationError(message)


class BaseUniqueFieldValidator(BaseValidator):
    """
    Validator that corresponds to `unique=True` on a model field.
    Should be applied to an individual field on the serializer.
    """
    message = u"This field must be unique."

    def __init__(self, queryset, message=None):
        super(BaseUniqueFieldValidator, self).__init__(message=self.message)
        self.queryset = queryset
        self.serializer_field = None
        self.message = message or self.message

    def set_context(self, serializer_field):
        """
        This hook is called by the serializer instance,
        prior to the validation call being made.
        """
        self.field_name = serializer_field.source_attrs[0]
        # Determine the existing instance, if this is an update operation.
        self.instance = getattr(serializer_field.parent, 'instance', None)

    def filter_queryset(self, value, queryset):
        raise NotImplementedError('`filter_queryset()` must be implemented.')

    def exclude_current_instance(self, queryset):
        """
        If an instance is being updated, then do not include
        that instance itself as a uniqueness conflict.
        """
        raise NotImplementedError('`exclude_current_instance()` must be '
                                  'implemented.')

    def __call__(self, attrs):
        raise NotImplementedError('`__call__` must be implemented.')

    def __repr__(self):
        return "<%s(queryset=%s)>" % (
            self.__class__.__name__,
            smart_repr(self.queryset)
        )



# -*- coding: utf-8 -*-
"""
Serializers for generated responses by the server.
"""
import json

from io import StringIO
from aiorest_ws.conf import settings
from aiorest_ws.exceptions import SerializerError
from aiorest_ws.utils.formatting import SHORT_SEPARATORS, LONG_SEPARATORS, \
    WRONG_UNICODE_SYMBOLS
from aiorest_ws.utils.xmlutils import SimpleXMLGenerator

__all__ = ('BaseRenderer', 'JSONRenderer', 'XMLRenderer', )


class BaseRenderer(object):

    format = None
    charset = 'utf-8'

    def render(self, data):
        """
        Render input data into another format.

        :param data: dictionary object.
        """
        pass


class JSONRenderer(BaseRenderer):

    format = 'json'
    # Don't set a charset because JSON is a binary encoding, that can be
    # encoded as utf-8, utf-16 or utf-32.
    # For more details see: http://www.ietf.org/rfc/rfc4627.txt
    # and Armin Ronacher's article http://goo.gl/MExCKv
    charset = None
    ensure_ascii = not settings.UNICODE_JSON
    compact = settings.COMPACT_JSON

    def render(self, data):
        """
        Render input data into JSON.

        :param data: dictionary or list object (response).
        """
        separators = SHORT_SEPARATORS if self.compact else LONG_SEPARATORS

        try:
            render = json.dumps(
                data, ensure_ascii=self.ensure_ascii, separators=separators
            )

            # Unicode symbols \u2028 and \u2029 are invisible in JSON and
            # make output are invalid. To avoid this situations, necessary
            # replace this symbols.
            # For more information read this article: http://goo.gl/ImC89E
            for wrong_symbol, expected in WRONG_UNICODE_SYMBOLS:
                render = render.replace(wrong_symbol, expected)

            render = bytes(render.encode('utf-8'))
        except Exception as exc:
            raise SerializerError(exc)
        return render


class XMLRenderer(BaseRenderer):

    format = 'xml'
    xml_generator = SimpleXMLGenerator

    def render(self, data):
        """
        Render input data into XML.

        :param data: dictionary or list object (response).
        """
        try:
            render = StringIO()
            xml = self.xml_generator(render, self.charset)
            xml.parse(data)
            render = bytes(render.getvalue().encode('utf-8'))
        except Exception as exc:
            raise SerializerError(exc)
        return render



# -*- coding: utf-8 -*-
"""
Exception classes for token authentication.
"""
from aiorest_ws.exceptions import BaseAPIException

__all__ = (
    'BaseTokenException', 'ParsingTokenException', 'InvalidSignatureException',
    'TokenNotBeforeException', 'TokenExpiredException',
    'TokenNotProvidedException',
)


class BaseTokenException(BaseAPIException):
    default_detail = u"Occurred error, when aiorest-ws processing token."


class ParsingTokenException(BaseTokenException):
    default_detail = u"Can't parse token: mismatch format."


class InvalidSignatureException(BaseTokenException):
    default_detail = u"Token signature is invalid."


class TokenNotBeforeException(BaseTokenException):
    default_detail = u"Time after which the token not be accepted has passed."


class TokenExpiredException(BaseTokenException):
    default_detail = u"Token has expired."


class TokenNotProvidedException(BaseTokenException):
    default_detail = u"Token has not provided in the request."



# -*- coding: utf-8 -*-
"""
Module with special functions, which using for extracting data from inner
`Meta` table of certain model serializer.
"""
from collections import OrderedDict
from sqlalchemy.orm.interfaces import MANYTOMANY, MANYTOONE, ONETOMANY

from aiorest_ws.utils.structures import FieldInfo, RelationInfo

__all__ = [
    '_get_pk', '_get_fields', '_get_to_field', '_get_relations',
    '_get_forward_relationships', '_get_reverse_relationships',
    '_merge_fields_and_pk', '_merge_relationships', 'get_field_info',
    'is_abstract_model', 'get_relations_data', 'model_pk'
]


def _get_pk(mapper):
    """
    Extract primary keys from passed table.
    Mapper of SQLAlchemy returns list of primary keys by default.
    """
    return [(column.name, column) for column in mapper.primary_key]


def _get_fields(mapper):
    """
    Extract list of available fields in the table.
    """
    return OrderedDict(mapper.columns.items())


def _get_relations(mapper, direction_symbol):
    """
    Return a 'RelationshipProperty' properties maintained by mapper.
    """
    return [
        (field_name, relation)
        for field_name, relation in mapper.relationships.items()
        if relation.direction == direction_symbol
    ]


def _get_to_field(field):
    """
    Extract remote field with which model had relation.
    """
    remote_side = getattr(field, 'remote_side', None)
    return list(remote_side)[0] if remote_side else None


def _get_forward_relationships(mapper):
    """
    Returns an `OrderedDict` of field names to `RelationInfo`.
    """
    forward_relations = OrderedDict()

    # Foreign keys (one to many)
    fk_relations = _get_relations(mapper, ONETOMANY)
    for field_name, relationship_property in fk_relations:
        forward_relations[field_name] = RelationInfo(
            model_field=relationship_property,
            related_model=mapper.class_,
            to_many=False,
            to_field=_get_to_field(relationship_property),
            has_through_model=False
        )

    return forward_relations


def _get_reverse_relationships(mapper):
    """
    Returns an `OrderedDict` of field names to `RelationInfo`.

    """
    reverse_relations = OrderedDict()

    # Foreign keys (many to one)
    reverse_fk_relations = _get_relations(mapper, MANYTOONE)
    for field_name, relationship_property in reverse_fk_relations:
        reverse_relations[field_name] = RelationInfo(
            model_field=relationship_property,
            related_model=mapper.class_,
            to_many=False,
            to_field=_get_to_field(relationship_property),
            has_through_model=False
        )

    # Many to many
    m2m_relations = _get_relations(mapper, MANYTOMANY)
    for field_name, relationship_property in m2m_relations:
        reverse_relations[field_name] = RelationInfo(
            model_field=relationship_property,
            related_model=mapper.class_,
            to_many=True,
            # manytomany do not have to_fields
            to_field=None,
            has_through_model=True
        )

    return reverse_relations


def _merge_fields_and_pk(pk, fields):
    """
    Collecting all fields into one dictionary object.
    """
    fields_and_pk = OrderedDict()
    fields_and_pk.update(pk)
    fields_and_pk.update(fields)
    return fields_and_pk


def _merge_relationships(forward_relations, reverse_relations):
    """
    Merge two different relations into one dictionary object.
    """
    return OrderedDict(
        list(forward_relations.items()) + list(reverse_relations.items())
    )


def get_field_info(model):
    """
    Given a model class, returns a `FieldInfo` instance, which is a
    `namedtuple`, containing metadata about the various field types on the
    model including information about their relationships.
    """
    mapper = model.__mapper__

    pk = _get_pk(mapper)
    fields = _get_fields(mapper)
    forward_relations = _get_forward_relationships(mapper)
    reverse_relations = _get_reverse_relationships(mapper)
    fields_and_pk = _merge_fields_and_pk(pk, fields)
    relationships = _merge_relationships(forward_relations, reverse_relations)

    return FieldInfo(pk, fields, forward_relations, reverse_relations,
                     fields_and_pk, relationships)


def is_abstract_model(model):
    """
    Checks model whether it abstract or not.
    """
    is_instantiated_table = hasattr(model, '__table__')
    has_specified_abstract_attribute = getattr(model, '__abstract__', False)
    return not is_instantiated_table and has_specified_abstract_attribute


def get_relations_data(model, validated_data):
    relations = {}
    info = get_field_info(model)
    for field_name, relation_info in info.relations.items():
        is_relation_field = False
        if relation_info.to_many or relation_info.to_field is not None:
            is_relation_field = True

        if is_relation_field and field_name in validated_data:
            relations[field_name] = validated_data.pop(field_name)

    return relations


def model_pk(model):
    field_name_and_column_pairs = _get_pk(model.__mapper__)
    field_mapper = dict(field_name_and_column_pairs)
    return list(field_mapper.keys())



# -*- coding: utf-8 -*-
"""
Field entities, implemented for support Django ORM.

Every class, represented here, is associated with one certain field type of
table relatively to Django ORM. Each of them field also used later for
serializing/deserializing object of ORM.
"""
import datetime
import re
import uuid

from aiorest_ws.conf import settings
from aiorest_ws.db.orm import fields
from aiorest_ws.db.orm.django.compat import get_remote_field, \
    value_from_object
from aiorest_ws.db.orm.fields import empty
from aiorest_ws.db.orm.validators import MaxLengthValidator
from aiorest_ws.utils.date.dateparse import parse_duration

from django.forms import FilePathField as DjangoFilePathField
from django.forms import ImageField as DjangoImageField
from django.core.exceptions import ValidationError as DjangoValidationError
from django.core.validators import EmailValidator, RegexValidator, \
    URLValidator, ip_address_validators
from django.utils import six
from django.utils.duration import duration_string
from django.utils.encoding import is_protected_type
from django.utils.ipv6 import clean_ipv6_address

__all__ = (
    'IntegerField', 'BooleanField', 'CharField', 'ChoiceField',
    'MultipleChoiceField', 'FloatField', 'NullBooleanField', 'DecimalField',
    'TimeField', 'DateField', 'DateTimeField', 'DurationField', 'ListField',
    'DictField', 'HStoreField', 'JSONField', 'ModelField', 'ReadOnlyField',
    'SerializerMethodField', 'EmailField', 'RegexField', 'SlugField',
    'URLField', 'UUIDField', 'IPAddressField', 'FilePathField', 'FileField',
    'ImageField', 'CreateOnlyDefault'
)


class IntegerField(fields.IntegerField):
    pass


class BooleanField(fields.BooleanField):
    pass


class CharField(fields.CharField):
    pass


class ChoiceField(fields.ChoiceField):
    pass


class MultipleChoiceField(ChoiceField):
    default_error_messages = {
        'invalid_choice': u'"{input}" is not a valid choice.',
        'not_a_list': u'Expected a list of items but got type "{input_type}".',
        'empty': u'This selection may not be empty.'
    }

    def __init__(self, *args, **kwargs):
        self.allow_empty = kwargs.pop('allow_empty', True)
        super(MultipleChoiceField, self).__init__(*args, **kwargs)

    def get_value(self, dictionary):
        if self.field_name not in dictionary:
            if getattr(self.root, 'partial', False):
                return empty
        return dictionary.get(self.field_name, empty)

    def to_internal_value(self, data):
        if isinstance(data, type('')) or not hasattr(data, '__iter__'):
            self.raise_error('not_a_list', input_type=type(data).__name__)
        if not self.allow_empty and len(data) == 0:
            self.raise_error('empty')

        return {
            super(MultipleChoiceField, self).to_internal_value(item)
            for item in data
        }

    def to_representation(self, value):
        return {
            self.choice_strings_to_values.get(str(item), item)
            for item in value
        }


class FloatField(fields.FloatField):
    pass


class NullBooleanField(fields.NullBooleanField):
    pass


class DecimalField(fields.DecimalField):
    pass


class TimeField(fields.TimeField):
    pass


class DateField(fields.DateField):
    pass


class DateTimeField(fields.DateTimeField):
    pass


class DurationField(fields.AbstractField):
    default_error_messages = {
        'invalid': u"Duration has wrong format. Use one of these formats "
                   u"instead: {format}.",
    }

    def to_internal_value(self, value):
        if isinstance(value, datetime.timedelta):
            return value
        parsed = parse_duration(str(value))
        if parsed is not None:
            return parsed
        self.raise_error('invalid', format='[DD] [HH:[MM:]]ss[.uuuuuu]')

    def to_representation(self, value):
        return duration_string(value)


class ListField(fields.ListField):
    pass


class DictField(fields.DictField):
    pass


class HStoreField(fields.HStoreField):
    pass


class JSONField(fields.JSONField):
    pass


class ModelField(fields.ModelField):
    default_error_messages = {
        'max_length': u'Ensure this field has no more than {max_length} '
                      u'characters.'
    }

    def __init__(self, model_field, **kwargs):
        # The `max_length` option is supported by Django's base `Field` class,
        # so we'd better support it here
        max_length = kwargs.pop('max_length', None)
        super(ModelField, self).__init__(model_field, **kwargs)
        if max_length is not None:
            message = self.error_messages['max_length'].format(max_length=max_length)  # NOQA
            self.validators.append(MaxLengthValidator(max_length, message=message))  # NOQA

    def to_internal_value(self, data):
        rel = get_remote_field(self.model_field, default=None)
        if rel is not None:
            return rel.to._meta.get_field(rel.field_name).to_python(data)
        return self.model_field.to_python(data)

    def to_representation(self, obj):
        value = value_from_object(self.model_field, obj)
        if is_protected_type(value):
            return value
        return self.model_field.value_to_string(obj)


class ReadOnlyField(fields.ReadOnlyField):
    pass


class SerializerMethodField(fields.SerializerMethodField):
    pass


class CreateOnlyDefault(fields.CreateOnlyDefault):
    pass


class EmailField(CharField):
    default_error_messages = {
        "invalid": u"Enter a valid email address."
    }

    def __init__(self, **kwargs):
        super(EmailField, self).__init__(**kwargs)
        validator = EmailValidator(message=self.error_messages['invalid'])
        self.validators.append(validator)


class RegexField(CharField):
    default_error_messages = {
        'invalid': u"This value does not match the required pattern."
    }

    def __init__(self, regex, **kwargs):
        super(RegexField, self).__init__(**kwargs)
        validator = RegexValidator(
            regex, message=self.error_messages['invalid']
        )
        self.validators.append(validator)


class SlugField(CharField):
    default_error_messages = {
        'invalid': u'Enter a valid "slug" consisting of letters, numbers, '
                   u'underscores or hyphens.'
    }

    def __init__(self, **kwargs):
        super(SlugField, self).__init__(**kwargs)
        slug_regex = re.compile(r'^[-a-zA-Z0-9_]+$')
        validator = RegexValidator(
            slug_regex, message=self.error_messages['invalid']
        )
        self.validators.append(validator)


class URLField(CharField):
    default_error_messages = {
        'invalid': u"Enter a valid URL."
    }

    def __init__(self, **kwargs):
        super(URLField, self).__init__(**kwargs)
        validator = URLValidator(message=self.error_messages['invalid'])
        self.validators.append(validator)


class UUIDField(fields.AbstractField):
    valid_formats = ('hex_verbose', 'hex', 'int', 'urn')

    default_error_messages = {
        'invalid': u'"{value}" is not a valid UUID.'
    }

    def __init__(self, **kwargs):
        self.uuid_format = kwargs.pop('format', 'hex_verbose')
        if self.uuid_format not in self.valid_formats:
            raise ValueError(
                'Invalid format for uuid representation. '
                'Must be one of "{0}"'.format('", "'.join(self.valid_formats))
            )
        super(UUIDField, self).__init__(**kwargs)

    def to_internal_value(self, data):
        if not isinstance(data, uuid.UUID):
            try:
                if isinstance(data, int):
                    return uuid.UUID(int=data)
                elif isinstance(data, str):
                    return uuid.UUID(hex=data)
                else:
                    self.raise_error('invalid', value=data)
            except ValueError:
                self.raise_error('invalid', value=data)
        return data

    def to_representation(self, value):
        if self.uuid_format == 'hex_verbose':
            return str(value)
        else:
            return getattr(value, self.uuid_format)


class IPAddressField(CharField):
    """Support both IPAddressField and GenericIPAddressField"""

    default_error_messages = {
        'invalid': u"Enter a valid IPv4 or IPv6 address."
    }

    def __init__(self, protocol='both', **kwargs):
        self.protocol = protocol.lower()
        self.unpack_ipv4 = (self.protocol == 'both')
        super(IPAddressField, self).__init__(**kwargs)
        validators, error_message = ip_address_validators(protocol, self.unpack_ipv4)  # NOQA
        self.validators.extend(validators)

    def to_internal_value(self, data):
        if not isinstance(data, six.string_types):
            self.raise_error('invalid', value=data)

        if ':' in data:
            try:
                if self.protocol in ('both', 'ipv6'):
                    return clean_ipv6_address(data, self.unpack_ipv4)
            except DjangoValidationError:
                self.raise_error('invalid', value=data)

        return super(IPAddressField, self).to_internal_value(data)


class FilePathField(ChoiceField):
    default_error_messages = {
        'invalid_choice': u'"{input}" is not a valid path choice.'
    }

    def __init__(self, path, match=None, recursive=False, allow_files=True,
                 allow_folders=False, required=None, **kwargs):
        # Defer to Django's FilePathField implementation to get the
        # valid set of choices
        field = DjangoFilePathField(
            path, match=match, recursive=recursive, allow_files=allow_files,
            allow_folders=allow_folders, required=required
        )
        kwargs['choices'] = field.choices
        super(FilePathField, self).__init__(**kwargs)


class FileField(fields.AbstractField):
    default_error_messages = {
        'required': u'No file was submitted.',
        'invalid': u'The submitted data was not a file. Check the encoding '
                   u'type on the form.',
        'no_name': u'No filename could be determined.',
        'empty': u'The submitted file is empty.',
        'max_length': u'Ensure this filename has at most {max_length} '
                      u'characters (it has {length}).',
    }

    def __init__(self, *args, **kwargs):
        self.max_length = kwargs.pop('max_length', None)
        self.allow_empty_file = kwargs.pop('allow_empty_file', False)
        if 'use_url' in kwargs:
            self.use_url = kwargs.pop('use_url')
        super(FileField, self).__init__(*args, **kwargs)

    def to_internal_value(self, data):
        try:
            # `UploadedFile` objects should have name and size attributes
            file_name = data.name
            file_size = data.size
        except AttributeError:
            self.raise_error('invalid')

        if not file_name:
            self.raise_error('no_name')
        if not self.allow_empty_file and not file_size:
            self.raise_error('empty')
        if self.max_length and len(file_name) > self.max_length:
            self.raise_error(
                'max_length', max_length=self.max_length, length=len(file_name)
            )

        return data

    def to_representation(self, value):
        use_url = getattr(self, 'use_url', settings.UPLOADED_FILES_USE_URL)

        if not value:
            return None

        if use_url:
            if not getattr(value, 'url', None):
                # If the file has not been saved it may not have a URL
                return None
            url = value.url
            return url
        return value.name


class ImageField(FileField):
    default_error_messages = {
        'invalid_image': u'Upload a valid image. The file you uploaded was '
                         u'either not an image or a corrupted image.'
    }

    def __init__(self, *args, **kwargs):
        self._DjangoImageField = kwargs.pop(
            '_DjangoImageField', DjangoImageField
        )
        super(ImageField, self).__init__(*args, **kwargs)

    def to_internal_value(self, data):
        # Image validation is a bit grungy, so we'll just outright
        # defer to Django's implementation so we don't need to
        # consider it, or treat PIL as a test dependency
        file_object = super(ImageField, self).to_internal_value(data)
        django_field = self._DjangoImageField()
        django_field.error_messages = self.error_messages
        django_field.to_python(file_object)
        return file_object



# -*- coding: utf-8 -*-
"""
Base serialize classes for support serialize any ORM objects.

All this classes are based on serializers from Django REST framework. Original
code published under BSD license. For more details look into AUTHORS file.
"""
import copy
from abc import abstractmethod

from aiorest_ws.db.orm.exceptions import ValidationError
from aiorest_ws.utils.fields import get_attribute
from aiorest_ws.utils.functional import cached_property

__all__ = (
    'ERROR_MESSAGE_NOT_FOUND', 'SkipField', 'empty', 'AbstractSerializer',
    'AbstractField',
)

ERROR_MESSAGE_NOT_FOUND = (
    u'Key `{key}` not found in the `error_messages` dictionary.'
)


class empty(object):
    """
    This class is wrapper, which used to understand, passed to validators
    `empty` value or not.
    It is required, because we can take collisions in situations, when
    user passed `None` value to some function, which means that necessary
    to set `None` value for field of model.
    """
    pass


class SkipField(Exception):
    """
    This class used in situations, when necessary to skip specified field. For
    example, it can be partial updates.
    """
    pass


class AbstractSerializer(object):
    """
    Base abstract class for serialize ORM object to dictionary object.

    This class provide few methods, which will be helpful when implementing
    custom serializers for concrete ORM engine.
    """
    _creation_counter = 0
    default_error_messages = {
        'required': u"This field is required.",
        'null': u"This field doesn't accept null values."
    }
    default_validators = []
    initial = None

    def __init__(self, read_only=False, write_only=False, required=None,
                 default=empty, initial=empty, source=None, label=None,
                 error_messages=None, validators=None, allow_null=False):
        self._creation_counter = AbstractField._creation_counter
        AbstractField._creation_counter += 1

        self.read_only = read_only
        self.write_only = write_only
        self.required = required
        self.default = default
        self.source = source
        self.initial = self.initial if (initial is empty) else initial
        self.label = label
        self.allow_null = allow_null

        # Set `required` by default to `True` value, unless specified
        if self.required is None:
            self.required = self.default is empty and not self.read_only

        # Make a copy of validators for concrete field
        if validators is not None:
            self._validators = validators[:]

        # These attributes using in set up of `bind()` method, when field
        # is bonded with model serializer
        self.field_name = None
        self.parent = None

        # Build dictionary with errors, which based on the
        # `default_error_messages` attribute of every class in hierarchy
        messages = {}
        for cls in reversed(self.__class__.__mro__):
            messages.update(getattr(cls, 'default_error_messages', {}))
        messages.update(error_messages or {})
        self.error_messages = messages

    def __new__(cls, *args, **kwargs):
        """
        When a field is instantiated, we store the arguments that were used,
        so that we can present a helpful representation of the object.
        """
        instance = super(AbstractSerializer, cls).__new__(cls)
        instance._args = args
        instance._kwargs = kwargs
        return instance

    def __deepcopy__(self, memo):
        """
        When cloning fields we instantiate using the arguments it was
        originally created with, rather than copying the complete state.
        """
        args = copy.deepcopy(self._args)
        kwargs = dict(self._kwargs)
        validators = kwargs.pop('validators', None)
        kwargs = copy.deepcopy(kwargs)
        if validators is not None:
            kwargs['validators'] = validators
        return self.__class__(*args, **kwargs)

    def bind(self, field_name, parent):
        """
        Initializes the field name and parent for the field instance.
        Called when a field is added to the parent serializer instance.

        :param field_name: string, which describe field name of model.
        :param parent: object, which store this instance of class.
        """
        self.field_name = field_name
        self.parent = parent

        # `self.label` should default to being based on the field name
        if self.label is None:
            self.label = field_name.replace('_', ' ').capitalize()

        # `self.source` field using for getting value from object of ORM model
        if self.source is None:
            self.source = field_name

        # `self.source_attrs` is a list of attributes that need to be looked
        # up when serializing the instance, or populating the validated data
        if self.source == '*':
            self.source_attrs = []
        else:
            self.source_attrs = self.source.split('.')

    @cached_property
    def root(self):
        """
        Returns the top-level serializer for this field.
        """
        root = self
        while root.parent is not None:
            root = root.parent
        return root

    @cached_property
    def context(self):
        """
        Returns the context as passed to the root serializer on initialization.
        """
        return getattr(self.root, '_context', {})

    @property
    def validators(self):
        """
        Return list of used validators for specified field.
        """
        if not hasattr(self, '_validators'):
            self._validators = self.get_validators()
        return self._validators

    @validators.setter
    def validators(self, validators):
        """
        Setter for validator attribute.

        :param validators: list of validators, which necessary to use.
        """
        self._validators = validators

    @abstractmethod
    def get_default(self):
        """
        Returns the default value when validating data not specified.
        """
        pass

    @abstractmethod
    def get_value(self, dictionary):
        """
        Returns the value from passed incoming data (dictionary object), which
        will be validated and converted to a native value.

        :param dictionary: dictionary, which represented as serialized object.
        """
        pass

    @abstractmethod
    def get_attribute(self, instance):
        """
        Given the *outgoing* object instance, return the primitive value
        that should be used for this field.

        :param instance: object, from which value will have taken.
        """
        pass

    @abstractmethod
    def to_internal_value(self, data):
        """
        Convert primitive data to native value.

        :param data: object, which necessary to convert to native value.
        """
        pass

    @abstractmethod
    def to_representation(self, value):
        """
        Convert the native value into primitive data.

        :param value: object, which necessary to convert into primitive data.
        """
        pass

    def get_validators(self):
        """
        Return a list copy of validators for current field.
        """
        return self.default_validators[:]

    def get_initial(self):
        """
        Return a value to use when the field is being returned as a primitive
        value, without any object instance.
        """
        return self.initial

    def validate_empty_values(self, data):
        """
        Validate empty values, and either:
        * Raise `ValidationError`, indicating invalid data.
        * Raise `SkipField`, indicating that the field should be ignored for
          partial object update.
        * Return (True, data), indicating an empty value and not necessary
          to apply any further validation.
        * Return (False, data), indicating a result, when have taken data
          must be validated in future.

        :param data: object, required for processing.
        """
        if self.read_only:
            return True, self.get_default()

        if data is empty:
            if getattr(self.root, 'partial', False):
                raise SkipField()
            if self.required:
                self.raise_error('required')
            return True, self.get_default()

        if data is None:
            if not self.allow_null:
                self.raise_error('null')
            return True, None

        return False, data

    def run_validators(self, value):
        """
        Check the given value through all validators for the specified field.

        :param value: value, required for processing.
        """
        errors = []
        for validator in self.validators:
            try:
                validator(value)
            except ValidationError as exc:
                errors.extend(exc.detail)
        if errors:
            raise ValidationError(errors)

    def run_validation(self, data=empty):
        """
        Validate a simple representation and return the internal value.
        The provided data may be `empty` if no representation was included
        in the input.

        :param data: object, which necessary to validate.
        """
        is_empty_value, data = self.validate_empty_values(data)
        if is_empty_value:
            return data
        value = self.to_internal_value(data)
        self.run_validators(value)
        return value

    def raise_error(self, key, **kwargs):
        """
        Helper function, that raise validation error with taken message.

        :param key: key for dictionary, by which selected concrete message.
        :param kwargs: dictionary object, which store additional arguments.
        """
        try:
            msg = self.error_messages[key]
        except KeyError:
            msg = ERROR_MESSAGE_NOT_FOUND.format(key=key)
            raise AssertionError(msg)
        message = msg.format(**kwargs)
        raise ValidationError(message)


class AbstractField(AbstractSerializer):
    """
    Base abstract class for model field classes.
    """
    def get_default(self):
        """
        Return the default value to use when validating data if no input
        is provided for this field.

        If a default has not been set for this field then this will simply
        return `empty`, indicating that no value should be set in the validated
        data for this field.
        """
        if self.default is empty:
            raise SkipField()
        return self.default

    def get_value(self, dictionary):
        """
        Returns the value from passed incoming data (dictionary object), which
        will be validated and converted to a native value.

        :param dictionary: dictionary, which represented as serialized object.
        """
        return dictionary.get(self.field_name, empty)

    def get_attribute(self, instance):
        """
        Given the *outgoing* object instance, return the primitive value
        that should be used for this field.

        :param instance: object, from which value will have taken.
        """
        try:
            return get_attribute(instance, self.source_attrs)
        except (KeyError, AttributeError) as exc:
            if not self.required and self.default is empty:
                raise SkipField()
            msg = (
                'Got {exc_type} when attempting to get a value for field '
                '`{field}` on serializer `{serializer}`.\nThe serializer '
                'field might be named incorrectly and not match '
                'any attribute or key on the `{instance}` instance.\n'
                'Original exception text was: {exc}.'.format(
                    exc_type=type(exc).__name__,
                    field=self.field_name,
                    serializer=self.parent.__class__.__name__,
                    instance=instance.__class__.__name__,
                    exc=exc
                )
            )
            raise type(exc)(msg)

    def to_internal_value(self, data):
        """
        Convert primitive data to native value.

        :param value: object, which necessary to convert to native value.
        """
        raise NotImplementedError(
            '{cls}.to_internal_value() must be implemented.'.format(
                cls=self.__class__.__name__
            )
        )

    def to_representation(self, value):
        """
        Convert the native value into primitive data.

        :param value: object, which necessary to convert into primitive data.
        """
        raise NotImplementedError(
            '{cls}.to_representation() must be implemented for field '
            '{field_name}. If you do not need to support write operations '
            'you are able to use subclass `ReadOnlyField` instead.'.format(
                cls=self.__class__.__name__,
                field_name=self.field_name,
            )
        )



# -*- coding: utf-8 -*-
"""
Module, which provide classes and function for related and nested field.
"""
from collections import OrderedDict

from aiorest_ws.db.orm import relations
from aiorest_ws.exceptions import ImproperlyConfigured
from aiorest_ws.utils.fields import get_attribute, is_simple_callable

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Manager
from django.db.models.query import QuerySet
from django.utils.encoding import smart_text


__all__ = (
    'ManyRelatedField', 'RelatedField', 'StringRelatedField',
    'PrimaryKeyRelatedField', 'HyperlinkedRelatedField',
    'HyperlinkedIdentityField', 'SlugRelatedField'
)


class ManyRelatedField(relations.ManyRelatedField):
    """
    Relationships with `many=True` transparently get coerced into instead being
    a ManyRelatedField with a child relationship.
    The `ManyRelatedField` class is responsible for handling iterating through
    the values and passing each one to the child relationship.
    This class is treated as private API.
    You shouldn't generally need to be using this class directly yourself,
    and should instead simply set 'many=True' on the relationship.
    """

    def get_attribute(self, instance):
        # Can't have any relationships if not created
        if hasattr(instance, 'pk') and instance.pk is None:
            return []

        relationship = get_attribute(instance, self.source_attrs)
        if hasattr(relationship, 'all'):
            return relationship.all()
        return relationship


class RelatedField(relations.RelatedField):
    many_related_field = ManyRelatedField

    def get_queryset(self):
        queryset = self.queryset
        if isinstance(queryset, (QuerySet, Manager)):
            # Ensure queryset is re-evaluated whenever used.
            # Note that actually a `Manager` class may also be used as the
            # queryset argument. This occurs on ModelSerializer fields,
            # as it allows us to generate a more expressive 'repr' output
            # for the field.
            # Eg: 'MyRelationship(queryset=ExampleModel.objects.all())'
            queryset = queryset.all()
        return queryset

    def get_attribute(self, instance):
        if self.use_pk_only_optimization() and self.source_attrs:
            # Optimized case, return a mock object only containing the
            # pk attribute
            try:
                instance = get_attribute(instance, self.source_attrs[:-1])
                value = instance.serializable_value(self.source_attrs[-1])
                if is_simple_callable(value):
                    # Handle edge case where the relationship `source` argument
                    # points to a `get_relationship()` method on the model
                    value = value().pk
                return relations.PKOnlyObject(pk=value)
            except AttributeError:
                pass

        # Standard case, return the object instance.
        return get_attribute(instance, self.source_attrs)

    def get_choices(self, cutoff=None):
        queryset = self.get_queryset()
        if queryset is None:
            # Ensure that field.choices returns something sensible
            # even when accessed with a read-only field
            return {}

        if cutoff is not None:
            queryset = queryset[:cutoff]

        return OrderedDict([
            (self.to_representation(item), self.display_value(item))
            for item in queryset
        ])

    @property
    def choices(self):
        return self.get_choices()

    @property
    def grouped_choices(self):
        return self.choices

    def display_value(self, instance):
        return str(instance)


class StringRelatedField(relations.StringRelatedField, RelatedField):
    pass


class PrimaryKeyRelatedField(relations.PrimaryKeyRelatedField,
                             RelatedField):

    def to_internal_value(self, data):
        if self.pk_field is not None:
            data = self.pk_field.to_internal_value(data)
        try:
            return self.get_queryset().get(pk=data)
        except ObjectDoesNotExist:
            self.raise_error('does_not_exist', pk_value=data)
        except (TypeError, ValueError):
            self.raise_error('incorrect_type', data_type=type(data).__name__)

    def to_representation(self, value):
        if self.pk_field is not None:
            return self.pk_field.to_representation(value.pk)
        return value.pk


class HyperlinkedRelatedField(relations.HyperlinkedRelatedField,
                              RelatedField):
    lookup_field = 'pk'

    def use_pk_only_optimization(self):
        return self.lookup_field == 'pk'

    def get_object(self, view_name, view_args, view_kwargs):
        """
        Return the object corresponding to a matched URL.
        Takes the matched URL conf arguments, and should return an
        object instance, or raise an `ObjectDoesNotExist` exception.
        """
        try:
            lookup_value = view_kwargs[self.lookup_url_kwarg]
            lookup_kwargs = {self.lookup_field: lookup_value}
            return self.get_queryset().get(**lookup_kwargs)
        except ObjectDoesNotExist:
            self.raise_error('does_not_exist')
        except KeyError:
            raise ImproperlyConfigured(
                "Missing primary key in the endpoint path. For fixing it just "
                "specify for a requested endpoint URL with included "
                "`{field_name}` parameter in the path or override"
                "`lookup_url_kwarg` in the constructor for the concrete field."
                .format(field_name=self.lookup_url_kwarg)
            )
        except (TypeError, ValueError):
            self.raise_error(
                'incorrect_type', data_type=type(view_kwargs).__name__
            )

    def is_saved_in_database(self, obj):
        if not obj or not obj.pk:
            return False
        return True

    def get_lookup_value(self, obj):
        pk = getattr(obj, self.lookup_field)
        return pk if isinstance(pk, (tuple, list)) else (pk, )


class HyperlinkedIdentityField(relations.HyperlinkedIdentityField,
                               HyperlinkedRelatedField):
    pass


class SlugRelatedField(relations.SlugRelatedField, RelatedField):
    """
    A read-write field that represents the target of the relationship
    by a unique 'slug' attribute.
    """
    def to_internal_value(self, data):
        try:
            return self.get_queryset().get(**{self.slug_field: data})
        except ObjectDoesNotExist:
            self.raise_error(
                'does_not_exist', slug_name=self.slug_field,
                value=smart_text(data)
            )
        except (TypeError, ValueError):
            self.raise_error('invalid')

    def to_representation(self, obj):
        return getattr(obj, self.slug_field)



# -*- coding: utf-8 -*-
"""
This module contains base serializer classes for ORM models/objects. They
are similar to Forms and ModelForms from Django/Flask frameworks.
"""
import copy
from inspect import isclass
from collections import OrderedDict

from aiorest_ws.conf import settings
from aiorest_ws.exceptions import ImproperlyConfigured
from aiorest_ws.db.orm.abstract import AbstractSerializer, AbstractField, \
    empty, SkipField
from aiorest_ws.db.orm.fields import *  # NOQA
from aiorest_ws.db.orm.exceptions import ValidationError
from aiorest_ws.utils.fields import set_value, get_attribute
from aiorest_ws.utils.functional import cached_property
from aiorest_ws.utils.representation import serializer_repr, list_repr
from aiorest_ws.utils.serializer_helpers import ReturnDict, ReturnList, \
    BoundField, BindingDict, NestedBoundField

__all__ = (
    'BaseSerializer', 'SerializerMetaclass', 'Serializer',
    'raise_errors_on_nested_writes', 'ListSerializer', 'ModelSerializer',
    'HyperlinkedModelSerializerMixin', 'LIST_SERIALIZER_KWARGS',
    'VALID_FIELD_KWARGS', 'ALL_FIELDS',
)

LIST_SERIALIZER_KWARGS = (
    'read_only', 'write_only', 'required', 'default', 'initial', 'source',
    'label', 'help_text', 'style', 'error_messages', 'allow_empty',
    'instance', 'data', 'partial', 'context', 'allow_null'
)
VALID_FIELD_KWARGS = {
    'read_only', 'write_only', 'required', 'default', 'initial', 'source',
    'label', 'help_text', 'style', 'error_messages', 'validators',
    'allow_null', 'allow_blank', 'choices'
}
ALL_FIELDS = '__all__'


class BaseSerializer(AbstractSerializer):
    """
    The BaseSerializer class provides a minimal class which may be used
    for writing custom serializer implementations.

    Note that we strongly restrict the ordering of operations/properties
    that may be used on the serializer in order to enforce correct usage.

    In particular, if a `data=` argument is passed then:

    .is_valid() - Available.
    .initial_data - Available.
    .validated_data - Only available after calling `is_valid()`
    .errors - Only available after calling `is_valid()`
    .data - Only available after calling `is_valid()`

    If a `data=` argument is not passed then:

    .is_valid() - Not available.
    .initial_data - Not available.
    .validated_data - Not available.
    .errors - Not available.
    .data - Available.
    """
    default_list_serializer = None  # override to your ListSerializer

    def __init__(self, instance=None, data=empty, **kwargs):
        self.instance = instance
        if data is not empty:
            self.initial_data = data
        self.partial = kwargs.pop('partial', False)
        self._context = kwargs.pop('context', {})
        kwargs.pop('many', None)
        super(BaseSerializer, self).__init__(**kwargs)

    def __new__(cls, *args, **kwargs):
        # We override this method in order to automatically create
        # `ListSerializer` classes instead when `many=True` is set
        if kwargs.pop('many', False):
            return cls.many_init(*args, **kwargs)
        return super(BaseSerializer, cls).__new__(cls, *args, **kwargs)

    @classmethod
    def many_init(cls, *args, **kwargs):
        """
        This method implements the creation of a `ListSerializer` parent
        class when `many=True` is used. You can customize it if you need to
        control which keyword arguments are passed to the parent, and which
        are passed to the child.

        Note that we're over-cautious in passing most arguments to both parent
        and child classes in order to try to cover the general case. If you're
        overriding this method you'll probably want something much simpler, eg:

        @classmethod
        def many_init(cls, *args, **kwargs):
            kwargs['child'] = cls()
            return CustomListSerializer(*args, **kwargs)
        """
        allow_empty = kwargs.pop('allow_empty', None)
        child_serializer = cls(*args, **kwargs)
        list_kwargs = {
            'child': child_serializer,
        }
        if allow_empty is not None:
            list_kwargs['allow_empty'] = allow_empty
        list_kwargs.update({
            key: value for key, value in kwargs.items()
            if key in LIST_SERIALIZER_KWARGS
        })
        meta = getattr(cls, 'Meta', None)
        list_serializer_class = getattr(
            meta, 'list_serializer_class', cls.default_list_serializer
        )
        if list_serializer_class is None:
            raise ImproperlyConfigured(
                "You must define `default_list_serializer` attribute for "
                "{cls} class before using this serializer for list of "
                "objects.".format(cls=cls)
            )
        return list_serializer_class(*args, **list_kwargs)

    @property
    def errors(self):
        if not hasattr(self, '_errors'):
            msg = 'You must call `.is_valid()` before ' \
                  'accessing `.errors`.'
            raise AssertionError(msg)
        return self._errors

    @property
    def validated_data(self):
        if not hasattr(self, '_validated_data'):
            msg = 'You must call `.is_valid()` before ' \
                  'accessing `.validated_data`.'
            raise AssertionError(msg)
        return self._validated_data

    @property
    def data(self):
        if hasattr(self, 'initial_data') and not hasattr(self, '_validated_data'):  # NOQA
            msg = (
                'When a serializer is passed a `data` keyword argument you '
                'must call `.is_valid()` before attempting to access the '
                'serialized `.data` representation.\n'
                'You should either call `.is_valid()` first, '
                'or access `.initial_data` instead.'
            )
            raise AssertionError(msg)

        if not hasattr(self, '_data'):
            if self.instance is not None and not getattr(self, '_errors', None):  # NOQA
                self._data = self.to_representation(self.instance)
            elif hasattr(self, '_validated_data') and not getattr(self, '_errors', None):  # NOQA
                self._data = self.to_representation(self.validated_data)
            else:
                self._data = self.get_initial()
        return self._data

    def to_internal_value(self, data):
        raise NotImplementedError('`to_internal_value()` must be implemented.')

    def to_representation(self, instance):
        raise NotImplementedError('`to_representation()` must be implemented.')

    def update(self, instance, validated_data):
        raise NotImplementedError('`update()` must be implemented.')

    def create(self, validated_data):
        raise NotImplementedError('`create()` must be implemented.')

    def save(self, **kwargs):
        assert hasattr(self, '_errors'), (
            'You must call `.is_valid()` before calling `.save()`.'
        )

        assert not self.errors, (
            'You cannot call `.save()` on a serializer with invalid data.'
        )

        assert 'commit' not in kwargs, (
            "'commit' is not a valid keyword argument to the 'save()' method. "
            "If you need to access data before committing to the database "
            "then inspect 'serializer.validated_data' instead. You can also "
            "pass additional keyword arguments to 'save()' if you need to "
            "set extra attributes on the saved model instance. "
            "For example: 'serializer.save(owner=request.user)'.'"
        )

        assert not hasattr(self, '_data'), (
            "You cannot call `.save()` after accessing `serializer.data`. If "
            "you need to access data before committing to the database then "
            "inspect 'serializer.validated_data' instead. "
        )

        validated_data = dict(
            list(self.validated_data.items()) +
            list(kwargs.items())
        )

        if self.instance is not None:
            self.instance = self.update(self.instance, validated_data)
            assert self.instance is not None, '`update()` did not return an object instance.'  # NOQA
        else:
            self.instance = self.create(validated_data)
            assert self.instance is not None, '`create()` did not return an object instance.'  # NOQA

        return self.instance

    def is_valid(self, raise_exception=False):
        assert hasattr(self, 'initial_data'), (
            'Cannot call `.is_valid()` as no `data=` keyword argument was '
            'passed when instantiating the serializer instance.'
        )

        if not hasattr(self, '_validated_data'):
            try:
                self._validated_data = self.run_validation(self.initial_data)
            except ValidationError as exc:
                self._validated_data = {}
                self._errors = exc.detail
            else:
                self._errors = {}

        if self._errors and raise_exception:
            raise ValidationError(self.errors)

        return not bool(self._errors)


class SerializerMetaclass(type):
    """
    This metaclass sets a dictionary named `_declared_fields` on the class.
    Any instances of `Field` included as attributes on either the class
    or on any of its superclasses will be include in the `_declared_fields`
    dictionary.
    """
    @classmethod
    def _get_declared_fields(cls, bases, attrs):
        fields = [
            (field_name, attrs.pop(field_name))
            for field_name, obj in list(attrs.items())
            if isinstance(obj, (AbstractField, BaseSerializer))
        ]
        fields.sort(key=lambda x: x[1]._creation_counter)

        # If this class is subclass of another Serializer, then add that
        # Serializer's fields. Note that we loop over the bases in *reverse*.
        # This is necessary in order to maintain the correct order of fields
        for base in reversed(bases):
            if hasattr(base, '_declared_fields'):
                fields = list(base._declared_fields.items()) + fields

        return OrderedDict(fields)

    def __new__(cls, name, bases, attrs):
        attrs['_declared_fields'] = cls._get_declared_fields(bases, attrs)
        return super(SerializerMetaclass, cls).__new__(cls, name, bases, attrs)


class Serializer(BaseSerializer, metaclass=SerializerMetaclass):
    default_error_messages = {
        'invalid': u"Invalid data. Expected a dictionary, but got {datatype}."
    }

    @property
    def fields(self):
        """
        A dictionary of {field_name: field_instance}.
        """
        # `fields` is evaluated lazily. We do this to ensure that we don't
        # have issues importing modules that use ModelSerializers as fields,
        # even if app-loading stage has not yet run
        if not hasattr(self, '_fields'):
            self._fields = BindingDict(self)
            for key, value in self.get_fields().items():
                self._fields[key] = value
        return self._fields

    @cached_property
    def _writable_fields(self):
        return [
            field for field in self.fields.values()
            if (not field.read_only) or (field.default is not empty)
        ]

    @cached_property
    def _readable_fields(self):
        return [
            field for field in self.fields.values()
            if not field.write_only
        ]

    def get_fields(self):
        """
        Returns a dictionary of {field_name: field_instance}.
        """
        # Every new serializer is created with a clone of the field instances.
        # This allows users to dynamically modify the fields on a serializer
        # instance without affecting every other serializer class
        return copy.deepcopy(self._declared_fields)

    def get_validators(self):
        """
        Returns a list of validator callables.
        """
        # Used by the lazily-evaluated `validators` property
        meta = getattr(self, 'Meta', None)
        validators = getattr(meta, 'validators', None)
        return validators[:] if validators else []

    def get_initial(self):
        if hasattr(self, 'initial_data'):
            return OrderedDict([
                (field_name, field.get_value(self.initial_data))
                for field_name, field in self.fields.items()
                if (field.get_value(self.initial_data) is not empty) and
                not field.read_only
            ])

        return OrderedDict([
            (field.field_name, field.get_initial())
            for field in self.fields.values()
            if not field.read_only
        ])

    def get_value(self, dictionary):
        return dictionary.get(self.field_name, empty)

    def run_validation(self, data=empty):
        """
        Validate passed data.
        """
        raise NotImplementedError('`run_validation()` must be implemented.')

    def to_internal_value(self, data):
        """
        Dict of native values <- Dict of primitive datatypes.
        """
        if not isinstance(data, dict):
            message = self.error_messages['invalid'].format(
                datatype=type(data).__name__
            )
            raise ValidationError({
                settings.REST_CONFIG['NON_FIELD_ERRORS_KEY']: [message]
            })

        ret = OrderedDict()
        errors = OrderedDict()
        fields = self._writable_fields

        for field in fields:
            validate_method = getattr(
                self, 'validate_' + field.field_name, None
            )
            primitive_value = field.get_value(data)
            try:
                validated_value = field.run_validation(primitive_value)
                if validate_method is not None:
                    validated_value = validate_method(validated_value)
            except ValidationError as exc:
                errors[field.field_name] = exc.detail
            except SkipField:
                pass
            else:
                set_value(ret, field.source_attrs, validated_value)

        if errors:
            raise ValidationError(errors)

        return ret

    def to_representation(self, instance):
        """
        Object instance -> Dict of primitive datatypes.
        """
        ret = OrderedDict()
        fields = self._readable_fields

        for field in fields:
            try:
                attribute = field.get_attribute(instance)
            except SkipField:
                continue

            if attribute is None:
                # We skip `to_representation` for `None` values so that
                # fields do not have to explicitly deal with that case
                ret[field.field_name] = None
            else:
                ret[field.field_name] = field.to_representation(attribute)

        return ret

    def validate(self, attrs):
        return attrs

    def __repr__(self):
        return serializer_repr(self, indent=1)

    # The following are used for accessing `BoundField` instances on the
    # serializer, for the purposes of presenting a form-like API onto the
    # field values and field errors.

    def __iter__(self):
        for field in self.fields.values():
            yield self[field.field_name]

    def __getitem__(self, key):
        field = self.fields[key]
        value = self.data.get(key)
        error = self.errors.get(key) if hasattr(self, '_errors') else None
        if isinstance(field, Serializer):
            return NestedBoundField(field, value, error)
        return BoundField(field, value, error)

    @property
    def data(self):
        ret = super(Serializer, self).data
        return ReturnDict(ret, serializer=self)

    @property
    def errors(self):
        ret = super(Serializer, self).errors
        return ReturnDict(ret, serializer=self)


def raise_errors_on_nested_writes(method_name, serializer, validated_data):
    """
    Give explicit errors when users attempt to pass writable nested data.
    If we don't do this explicitly they'd get a less helpful error when
    calling `.save()` on the serializer.
    We don't *automatically* support these sorts of nested writes because
    there are too many ambiguities to define a default behavior.
    Eg. Suppose we have a `UserSerializer` with a nested profile. How should
    we handle the case of an update, where the `profile` relationship does
    not exist? Any of the following might be valid:
    * Raise an application error.
    * Silently ignore the nested part of the update.
    * Automatically create a profile instance.
    """

    # Ensure we don't have a writable nested field. For example:
    #
    # class UserSerializer(ModelSerializer):
    #     ...
    #     profile = ProfileSerializer()
    assert not any(
        isinstance(field, BaseSerializer) and
        (key in validated_data) and
        isinstance(validated_data[key], (list, dict))
        for key, field in serializer.fields.items()
    ), (
        'The `.{method_name}()` method does not support writable nested '
        'fields by default.\nWrite an explicit `.{method_name}()` method for '
        'serializer `{module}.{class_name}`, or set `read_only=True` on '
        'nested serializer fields.'.format(
            method_name=method_name,
            module=serializer.__class__.__module__,
            class_name=serializer.__class__.__name__
        )
    )

    # Ensure we don't have a writable dotted-source field. For example:
    #
    # class UserSerializer(ModelSerializer):
    #     ...
    #     address = serializer.CharField('profile.address')
    assert not any(
        '.' in field.source and
        (key in validated_data) and
        isinstance(validated_data[key], (list, dict))
        for key, field in serializer.fields.items()
    ), (
        'The `.{method_name}()` method does not support writable '
        'dotted-source fields by default.\nWrite an explicit `.{method_name}'
        '()` method for serializer `{module}.{class_name}`, or set '
        '`read_only=True` on dotted-source serializer fields.'.format(
            method_name=method_name,
            module=serializer.__class__.__module__,
            class_name=serializer.__class__.__name__
        )
    )


class ListSerializer(BaseSerializer):
    child = None
    many = True

    default_error_messages = {
        'not_a_list': u'Expected a list of items but got type "{input_type}".',
        'empty': u"This list may not be empty."
    }

    def __init__(self, *args, **kwargs):
        self.child = kwargs.pop('child', copy.deepcopy(self.child))
        self.allow_empty = kwargs.pop('allow_empty', True)
        assert self.child is not None, '`child` is a required argument.'
        assert not isclass(self.child), '`child` has not been instantiated.'
        super(ListSerializer, self).__init__(*args, **kwargs)
        self.child.bind(field_name='', parent=self)

    def get_initial(self):
        if hasattr(self, 'initial_data'):
            return self.to_representation(self.initial_data)
        return []

    def get_value(self, dictionary):
        """
        Given the input dictionary, return the field value.
        """
        return dictionary.get(self.field_name, empty)

    def get_attribute(self, instance):
        try:
            return get_attribute(instance, self.source_attrs)
        except (KeyError, AttributeError) as exc:
            if not self.required and self.default is empty:
                raise SkipField()
            msg = (
                'Got {exc_type} when attempting to get a value for field '
                '`{field}` on serializer `{serializer}`.\nThe serializer '
                'field might be named incorrectly and not match '
                'any attribute or key on the `{instance}` instance.\n'
                'Original exception text was: {exc}.'.format(
                    exc_type=type(exc).__name__,
                    field=self.field_name,
                    serializer=self.parent.__class__.__name__,
                    instance=instance.__class__.__name__,
                    exc=exc
                )
            )
            raise type(exc)(msg)

    def run_validation(self, data=empty):
        """
        We override the default `run_validation`, because the validation
        performed by validators and the `.validate()` method should
        be coerced into an error dictionary with a 'non_fields_error' key.
        """
        raise NotImplementedError('`run_validation()` must be implemented.')

    def to_internal_value(self, data):
        """
        List of dicts of native values <- List of dicts of primitive datatypes.
        """
        if not isinstance(data, list):
            message = self.error_messages['not_a_list'].format(
                input_type=type(data).__name__
            )
            raise ValidationError({
                settings.REST_CONFIG['NON_FIELD_ERRORS_KEY']: [message]
            })

        if not self.allow_empty and len(data) == 0:
            message = self.error_messages['empty']
            raise ValidationError({
                settings.REST_CONFIG['NON_FIELD_ERRORS_KEY']: [message]
            })

        ret = []
        errors = []

        for item in data:
            try:
                validated = self.child.run_validation(item)
            except ValidationError as exc:
                errors.append(exc.detail)
            else:
                ret.append(validated)
                errors.append({})

        if any(errors):
            raise ValidationError(errors)

        return ret

    def to_representation(self, data):
        """
        List of object instances -> List of dicts of primitive datatypes.
        """
        # Dealing with nested relationships, data can be a Manager,
        # so, first get a queryset from the Manager if needed
        return [self.child.to_representation(item) for item in data]

    def validate(self, attrs):
        return attrs

    def update(self, instance, validated_data):
        raise NotImplementedError(
            "Serializers with many=True do not support multiple update by "
            "default, only multiple create. For updates it is unclear how to "
            "deal with insertions and deletions. If you need to support "
            "multiple update, use a `ListSerializer` class and override "
            "`.update()` so you can specify the behavior exactly."
        )

    def create(self, validated_data):
        return [self.child.create(attrs) for attrs in validated_data]

    def save(self, **kwargs):
        """
        Save and return a list of object instances.
        """
        # Guard against incorrect use of `serializer.save(commit=False)`
        assert 'commit' not in kwargs, (
            "'commit' is not a valid keyword argument to the 'save()' method. "
            "If you need to access data before committing to the database "
            "then inspect 'serializer.validated_data' instead. You can also "
            "pass additional keyword arguments to 'save()' if you need to "
            "set extra attributes on the saved model instance. For example: "
            "'serializer.save(owner=request.user)'.'"
        )

        validated_data = [
            dict(list(attrs.items()) + list(kwargs.items()))
            for attrs in self.validated_data
        ]

        if self.instance is not None:
            self.instance = self.update(self.instance, validated_data)
            assert self.instance is not None, '`update()` did not return an object instance.'  # NOQA
        else:
            self.instance = self.create(validated_data)
            assert self.instance is not None, '`create()` did not return an object instance.'  # NOQA

        return self.instance

    def __repr__(self):
        return list_repr(self, indent=1)

    @property
    def data(self):
        ret = super(ListSerializer, self).data
        return ReturnList(ret, serializer=self)

    @property
    def errors(self):
        ret = super(ListSerializer, self).errors
        if isinstance(ret, dict):
            return ReturnDict(ret, serializer=self)
        return ReturnList(ret, serializer=self)


class ModelSerializer(Serializer):
    """
    A `ModelSerializer` is just a regular `Serializer`, except that:
    * A set of default fields are automatically populated.
    * A set of default validators are automatically populated.
    * Default `.create()` and `.update()` implementations are provided.
    The process of automatically determining a set of serializer fields based
    on the model fields is reasonably complex, but you almost certainly don't
    need to dig into the implementation.
    If the `ModelSerializer` class *doesn't* generate the set of fields that
    you need you should either declare the extra/differing fields explicitly on
    the serializer class, or simply use a `Serializer` class.
    """
    serializer_field_mapping = {}  # override according with your ORM
    serializer_related_field = None  # override to your PrimaryKeyRelatedField
    serializer_related_to_field = None  # override to your SlugRelatedField
    serializer_url_field = None  # override to your HyperlinkedIdentityField
    serializer_choice_field = None  # override to your ChoiceField

    # The field name for hyperlinked identity fields. Defaults to 'url'.
    # You can modify this using the API setting.
    url_field_name = None

    def is_abstract_model(self, model):
        """
        Check the passed model is abstract.
        """
        raise NotImplementedError('`is_abstract_model()` must be implemented.')

    def get_field_info(self, model):
        """
        Get metadata about field in the passed model.
        """
        raise NotImplementedError('`get_field_info()` must be implemented.')

    # Default `create` and `update` behavior
    def create(self, validated_data):
        """
        Create object in the database, with the passed `validated_data`.
        """
        raise NotImplementedError('`create()` must be implemented.')

    def update(self, instance, validated_data):
        """
        Update existing object in database.
        """
        raise NotImplementedError('`create()` must be implemented.')

    def get_fields(self):
        """
        Return the dict of field names -> field instances that should be
        used for `self.fields` when instantiating the serializer.
        """
        if self.url_field_name is None:
            self.url_field_name = settings.REST_CONFIG['URL_FIELD_NAME']

        assert hasattr(self, 'Meta'), (
            'Class {serializer_class} missing "Meta" attribute'.format(
                serializer_class=self.__class__.__name__
            )
        )
        assert hasattr(self.Meta, 'model'), (
            'Class {serializer_class} missing "Meta.model" attribute'.format(
                serializer_class=self.__class__.__name__
            )
        )

        if self.is_abstract_model(self.Meta.model):
            raise ValueError('Cannot use ModelSerializer with Abstract Models.')  # NOQA

        declared_fields = copy.deepcopy(self._declared_fields)
        model = getattr(self.Meta, 'model')
        depth = getattr(self.Meta, 'depth', 0)

        if depth is not None:
            assert depth >= 0, "'depth' may not be negative."
            assert depth <= 10, "'depth' may not be greater than 10."

        # Retrieve metadata about fields & relationships on the model class
        info = self.get_field_info(model)
        field_names = self.get_field_names(declared_fields, info)

        # Determine any extra field arguments and hidden fields that
        # should be included
        extra_kwargs = self.get_extra_kwargs()
        extra_kwargs, hidden_fields = self.get_uniqueness_extra_kwargs(
            field_names, declared_fields, extra_kwargs
        )

        # Determine the fields that should be included on the serializer
        fields = OrderedDict()

        for field_name in field_names:
            # If the field is explicitly declared on the class then use that
            if field_name in declared_fields:
                fields[field_name] = declared_fields[field_name]
                continue

            # Determine the serializer field class and keyword arguments
            field_class, field_kwargs = self.build_field(
                field_name, info, model, depth
            )

            # Include any kwargs defined in `Meta.extra_kwargs`
            extra_field_kwargs = extra_kwargs.get(field_name, {})
            field_kwargs = self.include_extra_kwargs(
                field_kwargs, extra_field_kwargs
            )

            # Create the serializer field
            fields[field_name] = field_class(**field_kwargs)

        # Add in any hidden fields
        fields.update(hidden_fields)
        return fields

    def get_field_names(self, declared_fields, info):
        """
        Returns the list of all field names that should be created when
        instantiating this serializer class. This is based on the default
        set of fields, but also takes into account the `Meta.fields` or
        `Meta.exclude` options if they have been specified.
        """
        fields = getattr(self.Meta, 'fields', None)
        exclude = getattr(self.Meta, 'exclude', None)

        fields_are_specified = fields and fields != ALL_FIELDS
        if fields_are_specified and not isinstance(fields, (list, tuple)):
            raise TypeError(
                'The `fields` option must be a list or tuple or "__all__". '
                'Got %s.' % type(fields).__name__
            )

        if exclude and not isinstance(exclude, (list, tuple)):
            raise TypeError(
                'The `exclude` option must be a list or tuple. Got %s.' %
                type(exclude).__name__
            )

        assert not (fields and exclude), (
            "Cannot set both 'fields' and 'exclude' options on "
            "serializer {serializer_class}.".format(
                serializer_class=self.__class__.__name__
            )
        )

        if fields == ALL_FIELDS:
            fields = None

        if fields is not None:
            # Ensure that all declared fields have also been included in the
            # `Meta.fields` option

            # Do not require any fields that are declared a parent class,
            # in order to allow serializer subclasses to only include
            # a subset of fields
            required_field_names = set(declared_fields)
            for cls in self.__class__.__bases__:
                _declared_fields = getattr(cls, '_declared_fields', [])
                required_field_names -= set(_declared_fields)

            for field_name in required_field_names:
                assert field_name in fields, (
                    "The field '{field_name}' was declared on serializer "
                    "{serializer_class}, but has not been included in the "
                    "'fields' option.".format(
                        field_name=field_name,
                        serializer_class=self.__class__.__name__
                    )
                )

            return fields

        # Use the default set of field names if `Meta.fields` is not specified
        fields = self.get_default_field_names(declared_fields, info)

        if exclude is not None:
            # If `Meta.exclude` is included, then remove those fields
            for field_name in exclude:
                assert field_name in fields, (
                    "The field '{field_name}' was included on serializer "
                    "{serializer_class} in the 'exclude' option, but does "
                    "not match any model field.".format(
                        field_name=field_name,
                        serializer_class=self.__class__.__name__
                    )
                )
                fields.remove(field_name)

        return fields

    def get_default_field_names(self, declared_fields, model_info):
        """
        Return the default list of field names that will be used if the
        `Meta.fields` option is not specified.
        """
        raise NotImplementedError('`get_default_field_names()` must be '
                                  'implemented.')

    def build_field(self, field_name, info, model_class, nested_depth):
        """
        Return a two tuple of (cls, kwargs) to build a serializer field with.
        """
        raise NotImplementedError('`build_field()` must be implemented.')

    def build_standard_field(self, *args, **kwargs):
        """
        Create regular model fields.
        """
        raise NotImplementedError('`build_standard_field()` must be '
                                  'implemented.')

    def build_relational_field(self, *args, **kwargs):
        """
        Create fields for forward and reverse relationships.
        """
        raise NotImplementedError('`build_relational_field()` must be '
                                  'implemented.')

    def build_nested_field(self, *args, **kwargs):
        """
        Create nested fields for forward and reverse relationships.
        """
        raise NotImplementedError('`build_nested_field()` must be '
                                  'implemented.')

    def build_property_field(self, *args, **kwargs):
        """
        Create a read only field for model methods and properties.
        """
        raise NotImplementedError('`build_property_field()` must be '
                                  'implemented.')

    def build_url_field(self, *args, **kwargs):
        """
        Create a field representing the object's own URL.
        """
        raise NotImplementedError('`build_url_field()` must be implemented.')

    def build_unknown_field(self, field_name, model_class):
        """
        Raise an error on any unknown fields.
        """
        raise ImproperlyConfigured(
            'Field name `%s` is not valid for model `%s`.' %
            (field_name, model_class.__name__)
        )

    def include_extra_kwargs(self, kwargs, extra_kwargs):
        """
        Include any 'extra_kwargs' that have been included for this field,
        possibly removing any incompatible existing keyword arguments.
        """
        if extra_kwargs.get('read_only', False):
            for attr in [
                'required', 'default', 'allow_blank', 'allow_null',
                'min_length', 'max_length', 'min_value', 'max_value',
                'validators', 'queryset'
            ]:
                kwargs.pop(attr, None)

        if extra_kwargs.get('default') and kwargs.get('required') is False:
            kwargs.pop('required')

        if extra_kwargs.get('read_only', kwargs.get('read_only', False)):
            # Read only fields should always omit the 'required' argument
            extra_kwargs.pop('required', None)

        kwargs.update(extra_kwargs)

        return kwargs

    def get_extra_kwargs(self):
        """
        Return a dictionary mapping field names to a dictionary of
        additional keyword arguments.
        """
        extra_kwargs = copy.deepcopy(getattr(self.Meta, 'extra_kwargs', {}))

        read_only_fields = getattr(self.Meta, 'read_only_fields', None)
        if read_only_fields is not None:
            for field_name in read_only_fields:
                kwargs = extra_kwargs.get(field_name, {})
                kwargs['read_only'] = True
                extra_kwargs[field_name] = kwargs

        return extra_kwargs

    def get_uniqueness_extra_kwargs(self, field_names, declared_fields,
                                    extra_kwargs):
        """
        Return any additional field options that need to be included as a
        result of uniqueness constraints on the model. This is returned as
        a two-tuple of:
        ('dict of updated extra kwargs', 'mapping of hidden fields')
        """
        model = getattr(self.Meta, 'model')
        model_fields = self._get_model_fields(
            field_names, declared_fields, extra_kwargs
        )

        # Determine if we need any additional `HiddenField` or extra keyword
        # arguments to deal with `unique_for` dates that are required to
        # be in the input data in order to validate it
        unique_constraint_names = self._get_unique_constraint_names(
            model, model_fields, field_names
        )

        # Include each of "unique multiple columns" field names,
        # so long as all the field names are included on the serializer
        unique_constraint_names |= self._get_unique_together_constraints(
            model, model_fields, field_names
        )

        # Now we have all the field names that have uniqueness constraints
        # applied, we can add the extra 'required=...' or 'default=...'
        # arguments that are appropriate to these fields, or add a
        # `HiddenField` for it
        hidden_fields = {}
        uniqueness_extra_kwargs = {}

        for unique_constraint_name in unique_constraint_names:
            # Get the model field that is referred too
            unique_constraint_field = self._get_unique_field(
                model, unique_constraint_name
            )

            default = self._get_default_field_value(unique_constraint_field)

            if unique_constraint_name in model_fields:
                # The corresponding field is present in the serializer
                if default is empty:
                    field_kwargs = {'required': True}
                else:
                    field_kwargs = {'default': default}
                uniqueness_extra_kwargs[unique_constraint_name] = field_kwargs
            elif default is not empty:
                # The corresponding field is not present in the,
                # serializer. We have a default to use for it, so
                # add in a hidden field that populates it
                hidden_fields[unique_constraint_name] = HiddenField(default=default)  # NOQA

        # Update `extra_kwargs` with any new options
        for key, value in uniqueness_extra_kwargs.items():
            if key in extra_kwargs:
                extra_kwargs[key].update(value)
            else:
                extra_kwargs[key] = value

        return extra_kwargs, hidden_fields

    def _get_model_fields(self, field_names, declared_fields, extra_kwargs):
        """
        Returns all the model fields that are being mapped to by fields
        on the serializer class.
        Returned as a dict of 'model field name' -> 'model field'.
        Used internally by `get_uniqueness_field_options`.
        """
        model = getattr(self.Meta, 'model')
        model_fields = {}

        for field_name in field_names:
            if field_name in declared_fields:
                # If the field is declared on the serializer
                field = declared_fields[field_name]
                source = field.source or field_name
            else:
                try:
                    source = extra_kwargs[field_name]['source']
                except KeyError:
                    source = field_name

            if '.' in source or source == '*':
                # Model fields will always have a simple source mapping,
                # they can't be nested attribute lookups
                continue

            self._bind_field(model, source, model_fields)

        return model_fields

    def _get_unique_constraint_names(self, model, model_fields, field_names):
        """
        Return a set of field names, for each column unique constraint.
        Used internally by `get_uniqueness_extra_kwargs`.
        """
        raise NotImplementedError('`_get_unique_constraint_names()` '
                                  'must be implemented.')

    def _get_unique_together_constraints(self, model, model_fields, field_names):  # NOQA
        """
        Return a set of field names for a multiple unique constraints.
        Used internally by `get_uniqueness_extra_kwargs`.
        """
        raise NotImplementedError('`_get_unique_constraint_names()` '
                                  'must be implemented.')

    def _get_unique_field(self, model, unique_field_name):
        """
        Return a field by his name from a model.
        Used internally by `get_uniqueness_extra_kwargs`.
        """
        raise NotImplementedError('`_get_unique_field()` must be implemented.')

    def _get_default_field_value(self, unique_constraint_field):
        """
        Return a default value for a passed field.
        Used internally by `get_uniqueness_extra_kwargs`.
        """
        raise NotImplementedError('`_get_default_field_value()` '
                                  'must be implemented.')

    def _bind_field(self, model, source, model_fields):
        """
        Bind passed field to model serializer.
        Used internally by `_get_model_fields`.
        """
        raise NotImplementedError('`_bind_field()` must be implemented.')

    def get_validators(self):
        """
        Determine the set of validators to use when instantiating serializer.
        """
        meta = getattr(self, 'Meta', None)
        validators = getattr(meta, 'validators', None)
        return validators[:] if validators else []


class HyperlinkedModelSerializerMixin(object):
    """
    A type of `ModelSerializer` that uses hyperlinked relationships instead
    of primary key relationships. Specifically:
    * A 'url' field is included instead of the 'id' field.
    * Relationships to other instances are hyperlinks, instead of primary keys.
    """
    serializer_related_field = None  # override to your HyperlinkedRelatedField

    def get_default_field_names(self, declared_fields, model_info):
        """
        Return the default list of field names that will be used if the
        `Meta.fields` option is not specified.
        """
        return (
            [self.url_field_name, ] +
            list(declared_fields.keys()) +
            list(model_info.fields.keys()) +
            list(model_info.forward_relations.keys())
        )

    def build_nested_field(self, field_name, relation_info, nested_depth):
        """
        Create nested fields for forward and reverse relationships.
        """
        raise NotImplementedError('`build_nested_field()` must be '
                                  'implemented.')



# -*- coding: utf-8 -*-
"""
URL parsers, which help to define, with which endpoint router works.
"""
import re

from aiorest_ws.endpoints import PlainEndpoint, DynamicEndpoint
from aiorest_ws.exceptions import EndpointValueError

__all__ = (
    'ANY_VALUE', 'DYNAMIC_PARAMETER', 'VALID_DYNAMIC_PARAMETER', 'URLParser'
)

ANY_VALUE = r'[^{}/]+'
DYNAMIC_PARAMETER = re.compile(r'({\s*[\w\d_]+\s*})')
VALID_DYNAMIC_PARAMETER = re.compile(r'{(?P<var>[\w][\w\d_]*)}')


class URLParser(object):
    """
    Parser over endpoints paths, which returns one of the most suitable
    instances of route classes.
    """
    def define_route(self, path, handler, methods, name=None):
        """
        Define a router as instance of BaseRoute subclass, which passed
        from register method in RestWSRouter.

        :param path: URL, which used to get access to API.
        :param handler: class inherited from MethodBasedView, which used for
                        processing request.
        :param methods: list of available for user methods or string with
                        concrete method name.
        :param name: the base to use for the URL names that are created.
        """
        # it's PlainRoute, when don't have any "dynamic symbols"
        if all(symbol not in path for symbol in ['{', '}']):
            return PlainEndpoint(path, handler, methods, name)

        # Try to processing as a dynamic path
        pattern = ''
        for part in DYNAMIC_PARAMETER.split(path):
            match = VALID_DYNAMIC_PARAMETER.match(part)
            if match:
                pattern += '(?P<{}>{})'.format(match.group('var'), ANY_VALUE)
                continue

            if any(symbol in part for symbol in ['{', '}']):
                raise EndpointValueError("Invalid {} part of {} path".format(part, path))  # NOQA

            pattern += re.escape(part)
        try:
            compiled = re.compile("^{}$".format(pattern))
        except re.error as exc:
            raise EndpointValueError("Bad pattern '{}': {}".format(pattern, exc))  # NOQA
        return DynamicEndpoint(path, handler, methods, name, compiled)



# -*- coding: utf-8 -*-
"""
Token managers, proposed for generating/validating tokens.
"""
import hashlib
import hmac
import json
import time

from base64 import b64encode, b64decode
from aiorest_ws.auth.token.exceptions import ParsingTokenException, \
    InvalidSignatureException, TokenNotBeforeException, TokenExpiredException

__all__ = ('JSONWebTokenManager', )


class JSONWebTokenManager(object):
    """
    JSON Web Token (or shortly JWT) manager for the aiorest-ws library.

    This manager written under inspire of the articles below:
        https://scotch.io/tutorials/the-anatomy-of-a-json-web-token
        https://en.wikipedia.org/wiki/JSON_Web_Token
    """
    HASH_FUNCTIONS = {
        "HS256": hashlib.sha256,
        "HS384": hashlib.sha384,
        "HS512": hashlib.sha512
    }

    HASH_ALGORITHM = "HS256"
    SECRET_KEY = "secret_key"
    RESERVED_NAMES = ('iss', 'sub', 'aud', 'exp', 'nbf', 'ait', 'jti')

    def _encode_data(self, data):
        """
        Encode passed data to base64.

        :param data: dictionary object.
        """
        data = json.dumps(data).encode('utf-8')
        return b64encode(data).decode('utf-8')

    def _decode_data(self, data):
        """
        Decode passed data to JSON.

        :param data: dictionary object.
        """
        data = b64decode(data).decode('utf-8')
        return json.loads(data)

    def _generate_header(self):
        """
        Generate header for the token.
        """
        header = self._encode_data({"typ": "JWT", "alg": self.HASH_ALGORITHM})
        return header

    def _generate_payload(self, data):
        """
        Generate payload for the token.

        :param data: dictionary object.
        """
        payload = self._encode_data(data)
        return payload

    def _generate_signature(self, header, payload):
        """
        Generate signature for the token.

        :param header: token header.
        :param payload: token payload.
        """
        key = self.SECRET_KEY.encode('utf-8')
        data = "{0}.{1}".format(header, payload).encode('utf-8')
        hash_func = self.HASH_FUNCTIONS[self.HASH_ALGORITHM]

        hmac_obj = hmac.new(key, data, digestmod=hash_func)
        digest = hmac_obj.hexdigest().encode('utf-8')
        signature = b64encode(digest).decode('utf-8')
        return signature

    def _used_reserved_keys(self, data):
        """
        Get set of used reserved keys.
        """
        return set(data.keys()) & set(self.RESERVED_NAMES)

    def _check_token_timestamp(self, token, key):
        """
        Check token timestamp.

        :param token: dictionary object.
        :param key: field of token as a string.
        """
        token_timestamp = token.get(key, None)
        if token_timestamp:
            return time.time() > float(token_timestamp)
        return False

    def _is_invalid_signature(self, header, payload, token_signature):
        """
        Validate token by signature.

        :param header: header of token.
        :param payload: payload of token.
        :param token_signature: signature of token as a string.
        """
        server_signature = self._generate_signature(header, payload)
        if token_signature != server_signature:
            return True
        return False

    def _is_not_be_accepted(self, token):
        """
        Check for token is can be accepted or not.

        :param token: dictionary object.
        """
        return self._check_token_timestamp(token, 'nbf')

    def _is_expired_token(self, token):
        """
        Check for token expired or not.

        :param token: dictionary object.
        """
        return self._check_token_timestamp(token, 'exp')

    def set_reserved_attribute(self, token, attribute, value):
        """
        Set for token reserved attribute.

        :param token: dictionary object.
        :param attribute: updated reserved field of JSON Web Token.
        :param value: initialized value.
        """
        if attribute in self.RESERVED_NAMES and value:
            # If user define "exp" or "nbf" argument, than calculate timestamp
            if attribute in ['exp', 'nbf']:
                current_time_in_seconds = int(time.time())
                expired_timestamp = current_time_in_seconds + value
                token.update({attribute: expired_timestamp})
            # For any other JSON Web Token attributes just set value
            else:
                token[attribute] = value

    def generate(self, data, *args, **kwargs):
        """
        Generate token.

        :param data: dictionary, which will be stored inside token.
        :param args: tuple of arguments.
        :param kwargs: dictionary of reserved JSON Web Token fields, which
                       shall be overridden for token.
        """
        defined_attrs = self._used_reserved_keys(kwargs)
        for key in defined_attrs:
            self.set_reserved_attribute(data, key, kwargs[key])

        header = self._generate_header()
        payload = self._generate_payload(data)
        signature = self._generate_signature(header, payload)
        token = "{0}.{1}.{2}".format(header, payload, signature)
        return token

    def verify(self, token):
        """
        Verify passed token.

        :param token: validated token (as `header.payload.signature`).
        """
        try:
            header, payload, signature = token.split('.')
        except:
            raise ParsingTokenException()

        if self._is_invalid_signature(header, payload, signature):
            raise InvalidSignatureException()

        token_data = self._decode_data(payload)

        if self._is_not_be_accepted(token_data):
            raise TokenNotBeforeException()

        if self._is_expired_token(token_data):
            raise TokenExpiredException()

        return token_data



# -*- coding: utf-8 -*-
"""
List, model and hyperlinked serializer classes for Django ORM.

As you can see there, we are inherited from all base classes and implement
logic according to the work of SQLAlchemy ORM. For the most situations using
ModelSerializer class will be enough.
"""
import traceback

from aiorest_ws.conf import settings
from aiorest_ws.db.orm.abstract import empty
from aiorest_ws.db.orm.exceptions import ValidationError
from aiorest_ws.db.orm.serializers import \
    ListSerializer as BaseListSerializer, \
    ModelSerializer as BaseModelSerializer, HyperlinkedModelSerializerMixin, \
    raise_errors_on_nested_writes
from aiorest_ws.db.orm.django import model_meta
from aiorest_ws.db.orm.django.compat import postgres_fields, \
    JSONField as ModelJSONField
from aiorest_ws.db.orm.django.field_mapping import get_field_kwargs, \
    get_nested_relation_kwargs, get_relation_kwargs, get_url_kwargs
from aiorest_ws.utils.field_mapping import ClassLookupDict

from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import models
from django.db.models import DurationField as ModelDurationField
from django.db.models.fields import Field as DjangoModelField
from django.db.models.fields import FieldDoesNotExist
from django.utils import timezone

from aiorest_ws.db.orm.django.fields import *  # NOQA
from aiorest_ws.db.orm.django.relations import *  # NOQA

__all__ = (
    'get_validation_error_detail', 'ListSerializer', 'ModelSerializer',
    'HyperlinkedModelSerializer',
)


def get_validation_error_detail(exc):
    assert isinstance(exc, (ValidationError, DjangoValidationError))

    if isinstance(exc, DjangoValidationError):
        # Normally you should raise `serializers.ValidationError`
        # inside your codebase, but we handle Django's validation
        # exception class as well for simpler compat.
        # Eg. Calling Model.clean() explicitly inside Serializer.validate()
        return {settings.REST_CONFIG['NON_FIELD_ERRORS_KEY']: exc.args}
    elif isinstance(exc.detail, dict):
        # If errors may be a dict we use the standard {key: list of values}.
        # Here we ensure that all the values are *lists* of errors.
        return {
            key: value if isinstance(value, (list, dict)) else [value]
            for key, value in exc.detail.items()
        }
    elif isinstance(exc.detail, list):
        # Errors raised as a list are non-field errors.
        return {settings.REST_CONFIG['NON_FIELD_ERRORS_KEY']: exc.detail}
    # Errors raised as a string are non-field errors.
    return {settings.REST_CONFIG['NON_FIELD_ERRORS_KEY']: [exc.detail]}


class ListSerializer(BaseListSerializer):

    def run_validation(self, data=empty):
        """
            We override the default `run_validation`, because the validation
            performed by validators and the `.validate()` method should
            be coerced into an error dictionary with a 'non_fields_error' key.
            """
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data

        value = self.to_internal_value(data)
        try:
            self.run_validators(value)
            value = self.validate(value)
            assert value is not None, '.validate() should return the validated data'  # NOQA
        except (ValidationError, DjangoValidationError) as exc:
            raise ValidationError(detail=get_validation_error_detail(exc))

        return value


class ModelSerializer(BaseModelSerializer):
    """
    Base serializer for Django models.

    This class automatically generate "scheme" for further serializing and
    de-serializing data, according to used model, which is specified by user.
    For indicated fields by programmer, ModelSerializer will not apply "scheme
    construction rule". They are will have skipped while processing. Otherwise,
    for every field which is not pre-defined by user, will be found the most
    suitable and compatible field class.
    """
    serializer_field_mapping = {
        models.AutoField: IntegerField,
        models.BigIntegerField: IntegerField,
        models.BooleanField: BooleanField,
        models.CharField: CharField,
        models.CommaSeparatedIntegerField: CharField,
        models.DateField: DateField,
        models.DateTimeField: DateTimeField,
        models.DecimalField: DecimalField,
        models.EmailField: EmailField,
        models.Field: ModelField,
        models.FileField: FileField,
        models.FloatField: FloatField,
        models.ImageField: ImageField,
        models.IntegerField: IntegerField,
        models.NullBooleanField: NullBooleanField,
        models.PositiveIntegerField: IntegerField,
        models.PositiveSmallIntegerField: IntegerField,
        models.SlugField: SlugField,
        models.SmallIntegerField: IntegerField,
        models.TextField: CharField,
        models.TimeField: TimeField,
        models.URLField: URLField,
        models.GenericIPAddressField: IPAddressField,
        models.FilePathField: FilePathField,
    }
    if ModelDurationField is not None:
        serializer_field_mapping[ModelDurationField] = DurationField
    if ModelJSONField is not None:
        serializer_field_mapping[ModelJSONField] = JSONField
    serializer_related_field = PrimaryKeyRelatedField
    serializer_related_to_field = SlugRelatedField
    serializer_url_field = HyperlinkedIdentityField
    serializer_choice_field = ChoiceField
    default_list_serializer = ListSerializer

    def is_abstract_model(self, model):
        """
        Check the passed model is abstract.
        """
        return model_meta.is_abstract_model(model)

    def get_field_info(self, model):
        """
        Get metadata about field in the passed model.
        """
        return model_meta.get_field_info(model)

    # Default `create` and `update` behavior...
    def create(self, validated_data):
        """
        We have a bit of extra checking around this in order to provide
        descriptive messages when something goes wrong, but this method is
        essentially just:

            return ExampleModel.objects.create(**validated_data)

        If there are many to many fields present on the instance then they
        cannot be set until the model is instantiated, in which case the
        implementation is like so:

            example_relationship = validated_data.pop('example_relationship')
            instance = ExampleModel.objects.create(**validated_data)
            instance.example_relationship = example_relationship
            return instance

        The default implementation also does not handle nested relationships.
        If you want to support writable nested relationships you'll need
        to write an explicit `.create()` method.
        """
        raise_errors_on_nested_writes('create', self, validated_data)

        ModelClass = self.Meta.model

        # Remove many-to-many relationships from validated_data.
        # They are not valid arguments to the default `.create()` method,
        # as they require that the instance has already been saved
        info = model_meta.get_field_info(ModelClass)
        many_to_many = {}
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and (field_name in validated_data):
                many_to_many[field_name] = validated_data.pop(field_name)

        try:
            instance = ModelClass.objects.create(**validated_data)
        except TypeError:
            tb = traceback.format_exc()
            msg = (
                'Got a `TypeError` when calling `%s.objects.create()`. '
                'This may be because you have a writable field on the '
                'serializer class that is not a valid argument to '
                '`%s.objects.create()`. You may need to make the field '
                'read-only, or override the %s.create() method to handle '
                'this correctly.\nOriginal exception was:\n %s' %
                (
                    ModelClass.__name__,
                    ModelClass.__name__,
                    self.__class__.__name__,
                    tb
                )
            )
            raise TypeError(msg)

        # Save many-to-many relationships after the instance is created
        if many_to_many:
            for field_name, value in many_to_many.items():
                field = getattr(instance, field_name)
                field.add(*value)
            instance.save()

        return instance

    def update(self, instance, validated_data):
        raise_errors_on_nested_writes('update', self, validated_data)

        # Simply set each attribute on the instance, and then save it.
        # Note that unlike `.create()` we don't need to treat many-to-many
        # relationships as being a special case. During updates we already
        # have an instance pk for the relationships to be associated with
        for attr, value in validated_data.items():
            setattr(instance, attr, value)
        instance.save()

        return instance

    def run_validation(self, data=empty):
        """
        We override the default `run_validation`, because the validation
        performed by validators and the `.validate()` method should
        be coerced into an error dictionary with a 'non_fields_error' key.
        """
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data

        for field in self.fields.values():
            for validator in field.validators:
                if hasattr(validator, 'set_context'):
                    validator.set_context(field)

        value = self.to_internal_value(data)
        try:
            self.run_validators(value)
            value = self.validate(value)
            assert value is not None, '.validate() should return the ' \
                                      'validated data'
        except (ValidationError, DjangoValidationError) as exc:
            raise ValidationError(detail=get_validation_error_detail(exc))

        return value

    def get_default_field_names(self, declared_fields, model_info):
        """
        Return the default list of field names that will be used if the
        `Meta.fields` option is not specified.
        """
        return (
            [model_info.pk.name] +
            list(declared_fields.keys()) +
            list(model_info.fields.keys()) +
            list(model_info.forward_relations.keys())
        )

    def _get_unique_constraint_names(self, model, model_fields, field_names):
        """
        Return a set of field names, for each column unique constraint.
        Used internally by `get_uniqueness_extra_kwargs`.
        """
        unique_constraint_names = set()

        for model_field in model_fields.values():
            # Include each of the `unique_for_*` field names
            unique_constraint_names |= {
                model_field.unique_for_date, model_field.unique_for_month,
                model_field.unique_for_year
            }

        unique_constraint_names -= {None}
        return unique_constraint_names

    def _get_unique_together_constraints(self, model, model_fields, field_names):  # NOQA
        """
        Return a set of field names for a multiple unique constraints.
        Used internally by `get_uniqueness_extra_kwargs`.
        """
        unique_constraint_names = set()

        for parent_class in [model] + list(model._meta.parents.keys()):
            for unique_together_list in parent_class._meta.unique_together:
                if set(field_names).issuperset(set(unique_together_list)):
                    unique_constraint_names |= set(unique_together_list)

        return unique_constraint_names

    def _get_unique_field(self, model, unique_field_name):
        """
        Return a field by his name from a model.
        Used internally by `get_uniqueness_extra_kwargs`.
        """
        return model._meta.get_field(unique_field_name)

    def _get_default_field_value(self, unique_constraint_field):
        """
        Return a default value for a passed field.
        Used internally by `get_uniqueness_extra_kwargs`.
        """
        default = empty

        if getattr(unique_constraint_field, 'auto_now_add', None):
            default = CreateOnlyDefault(timezone.now)
        elif getattr(unique_constraint_field, 'auto_now', None):
            default = timezone.now
        elif unique_constraint_field.has_default():
            default = unique_constraint_field.default

        return default

    def build_field(self, field_name, info, model_class, nested_depth):
        """
        Create regular model fields.
        """
        if field_name in info.fields_and_pk:
            model_field = info.fields_and_pk[field_name]
            return self.build_standard_field(field_name, model_field)

        elif field_name in info.relations:
            relation_info = info.relations[field_name]
            if not nested_depth:
                return self.build_relational_field(field_name, relation_info)
            else:
                return self.build_nested_field(
                    field_name, relation_info, nested_depth
                )

        elif hasattr(model_class, field_name):
            return self.build_property_field(field_name, model_class)

        elif field_name == self.url_field_name:
            return self.build_url_field(field_name, model_class)

        return self.build_unknown_field(field_name, model_class)

    def build_standard_field(self, *args, **kwargs):
        """
        Create regular model fields.
        """
        field_name, model_field = args
        field_mapping = ClassLookupDict(self.serializer_field_mapping)

        field_class = field_mapping[model_field]
        field_kwargs = get_field_kwargs(field_name, model_field)

        if 'choices' in field_kwargs:
            # Fields with choices get coerced into `ChoiceField`
            # instead of using their regular typed field
            field_class = self.serializer_choice_field
            # Some model fields may introduce kwargs that would not be valid
            # for the choice field. We need to strip these out
            valid_kwargs = {
                'read_only', 'write_only',
                'required', 'default', 'initial', 'source',
                'label', 'help_text', 'style',
                'error_messages', 'validators', 'allow_null', 'allow_blank',
                'choices'
            }
            for key in list(field_kwargs.keys()):
                if key not in valid_kwargs:
                    field_kwargs.pop(key)

        if not issubclass(field_class, ModelField):
            # `model_field` is only valid for the fallback case of
            # `ModelField`, which is used when no other typed field
            # matched to the model field
            field_kwargs.pop('model_field', None)

        if not issubclass(field_class, CharField) and not issubclass(field_class, ChoiceField):  # NOQA
            # `allow_blank` is only valid for textual fields
            field_kwargs.pop('allow_blank', None)

        if postgres_fields and isinstance(model_field, postgres_fields.ArrayField):  # NOQA
            # Populate the `child` argument on `ListField` instances generated
            # for the PostgreSQL specific `ArrayField`
            child_model_field = model_field.base_field
            child_field_class, child_field_kwargs = self.build_standard_field(
                'child', child_model_field
            )
            field_kwargs['child'] = child_field_class(**child_field_kwargs)

        return field_class, field_kwargs

    def build_relational_field(self, *args, **kwargs):
        """
        Create fields for forward and reverse relationships.
        """
        field_name, relation_info = args
        field_class = self.serializer_related_field
        field_kwargs = get_relation_kwargs(field_name, relation_info)

        to_field = field_kwargs.pop('to_field', None)
        if to_field and not relation_info.related_model._meta.get_field(to_field).primary_key:  # NOQA
            field_kwargs['slug_field'] = to_field
            field_class = self.serializer_related_to_field

        # `view_name` is only valid for hyperlinked relationships
        if not issubclass(field_class, HyperlinkedRelatedField):
            field_kwargs.pop('view_name', None)

        return field_class, field_kwargs

    def build_nested_field(self, *args, **kwargs):
        """
        Create nested fields for forward and reverse relationships.
        """
        field_name, relation_info, nested_depth = args

        class NestedSerializer(ModelSerializer):

            class Meta:
                model = relation_info.related_model
                depth = nested_depth - 1

        field_class = NestedSerializer
        field_kwargs = get_nested_relation_kwargs(relation_info)

        return field_class, field_kwargs

    def build_property_field(self, *args, **kwargs):
        """
        Create a read only field for model methods and properties.
        """
        field_class = ReadOnlyField
        field_kwargs = {}

        return field_class, field_kwargs

    def build_url_field(self, field_name, model_class):
        """
        Create a field representing the object's own URL.
        """
        field_class = self.serializer_url_field
        field_kwargs = get_url_kwargs(model_class)

        return field_class, field_kwargs

    def _bind_field(self, model, source, model_fields):
        """
        Bind passed field to model serializer.
        Used internally by `_get_model_fields`.
        """
        try:
            field = model._meta.get_field(source)
            if isinstance(field, DjangoModelField):
                model_fields[source] = field
        except FieldDoesNotExist:
            pass


class HyperlinkedModelSerializer(HyperlinkedModelSerializerMixin,
                                 ModelSerializer):
    """
    A type of `ModelSerializer` that uses hyperlinked relationships instead
    of primary key relationships. Specifically:

    * A 'url' field is included instead of the 'id' field.
    * Relationships to other instances are hyperlinks, instead of primary keys.
    """
    serializer_related_field = HyperlinkedRelatedField

    def build_nested_field(self, field_name, relation_info, nested_depth):

        class NestedSerializer(HyperlinkedModelSerializer):

            class Meta:
                model = relation_info.related_model
                depth = nested_depth - 1

        field_class = NestedSerializer
        field_kwargs = get_nested_relation_kwargs(relation_info)

        return field_class, field_kwargs



# -*- coding: utf-8 -*-
"""
Exception classes for authentication.
"""
from aiorest_ws.exceptions import BaseAPIException

__all__ = ('BaseAuthException', 'PermissionDeniedException', )


class BaseAuthException(BaseAPIException):
    default_detail = u"Error occurred in the authentication process."


class PermissionDeniedException(BaseAuthException):
    default_detail = u"Permission denied: user doesn't have enough rights."
